from django.db import models


class CanvasUser(models.Model):
    """A class representing a Canvas User."""

    class Meta:
        managed = False
        app_label = "canvas_sdk"
        db_table = "canvas_sdk_data_api_auth_user_001"

    dbid = models.BigIntegerField(db_column="dbid", primary_key=True)



import shutil
from collections.abc import Generator
from datetime import datetime
from pathlib import Path
from typing import Any

import pytest
import typer
from typer.testing import CliRunner

from canvas_cli.main import app

from .plugin import validate_package

runner = CliRunner()


def test_validate_package_unexistant_path() -> None:
    """Tests the validate_package callback with an invalid folder."""
    with pytest.raises(typer.BadParameter):
        validate_package(Path("/a_random_url_that_will_not_exist_or_so_I_hope"))


def test_validate_package_wrong_file_type(tmp_path: Path) -> None:
    """Tests the validate_package callback with an invalid file type."""
    invalid_file = tmp_path / "tmp_file.zip"
    invalid_file.write_text("definitely not a python package")

    with pytest.raises(typer.BadParameter):
        validate_package(invalid_file)


def test_validate_package_valid_file(tmp_path: Path) -> None:
    """Tests the validate_package callback with a valid file type."""
    package_path = tmp_path / "test-package.whl"
    package_path.write_text("something")
    result = validate_package(package_path)
    assert result == package_path


@pytest.fixture(scope="session")
def init_plugin_name() -> str:
    """The plugin name to be used for the canvas cli init test."""
    return f"testing_init-{datetime.now().timestamp()}".replace(".", "")


@pytest.fixture(autouse=True, scope="session")
def clean_up_plugin(init_plugin_name: str) -> Generator[Any, Any, Any]:
    """Cleans up the plugin directory after the test."""
    yield
    if Path(f"./{init_plugin_name}").exists():
        shutil.rmtree(Path(f"./{init_plugin_name}"))


def test_canvas_init(init_plugin_name: str) -> None:
    """Tests that the CLI successfully creates a plugin with init."""
    result = runner.invoke(app, "init", input=init_plugin_name)
    assert result.exit_code == 0

    # plugin directory exists
    plugin = Path(f"./{init_plugin_name}")
    assert plugin.exists()
    assert plugin.is_dir()

    # manifest file exists
    manifest = Path(f"./{init_plugin_name}/CANVAS_MANIFEST.json")
    assert manifest.exists()
    assert manifest.is_file()
    manifest_result = runner.invoke(app, f"validate-manifest {init_plugin_name}")
    assert manifest_result.exit_code == 0

    # readme file exists
    readme = Path(f"./{init_plugin_name}/README.md")
    assert readme.exists()
    assert readme.is_file()

    # protocols dir exists
    protocols = Path(f"./{init_plugin_name}/protocols")
    assert protocols.exists()
    assert protocols.is_dir()

    # protocol file exists in protocols dir
    protocol = Path(f"./{init_plugin_name}/protocols/my_protocol.py")
    assert protocol.exists()
    assert protocol.is_file()



from django.db import models

from canvas_sdk.v1.data.base import CommittableModelManager
from canvas_sdk.v1.data.patient import Patient
from canvas_sdk.v1.data.user import CanvasUser


class IntervalUnit(models.TextChoices):
    """ProtocolOverride cycle IntervalUnit."""

    DAYS = "days", "days"
    MONTHS = "months", "months"
    YEARS = "years", "years"


class Status(models.TextChoices):
    """ProtocolOverride Status."""

    ACTIVE = "active", "active"
    INACTIVE = "inactive", "inactive"


class ProtocolOverride(models.Model):
    """ProtocolOverride."""

    class Meta:
        managed = False
        app_label = "canvas_sdk"
        db_table = "canvas_sdk_data_api_protocoloverride_001"

    objects = CommittableModelManager()

    id = models.UUIDField()
    dbid = models.BigIntegerField(primary_key=True)
    created = models.DateTimeField()
    modified = models.DateTimeField()
    deleted = models.BooleanField()
    committer = models.ForeignKey(CanvasUser, on_delete=models.DO_NOTHING)
    entered_in_error = models.ForeignKey(CanvasUser, on_delete=models.DO_NOTHING)
    patient = models.ForeignKey(
        Patient,
        on_delete=models.DO_NOTHING,
        related_name="protocol_overrides",
    )
    protocol_key = models.CharField()
    is_adjustment = models.BooleanField()
    reference_date = models.DateTimeField()
    cycle_in_days = models.IntegerField()
    is_snooze = models.BooleanField()
    snooze_date = models.DateField()
    snoozed_days = models.IntegerField()
    # reason_id = models.BigIntegerField()
    snooze_comment = models.TextField()
    narrative = models.CharField()
    # note_id = models.BigIntegerField()
    cycle_quantity = models.IntegerField()
    cycle_unit = models.CharField(choices=IntervalUnit.choices)
    status = models.CharField(choices=Status.choices)



from typing import Self

import arrow
from django.contrib.postgres.fields import ArrayField
from django.db import models
from django.db.models import TextChoices


class SexAtBirth(TextChoices):
    """SexAtBirth."""

    FEMALE = "F", "female"
    MALE = "M", "male"
    OTHER = "O", "other"
    UNKNOWN = "UNK", "unknown"
    BLANK = "", ""


class Patient(models.Model):
    """A class representing a patient."""

    class Meta:
        managed = False
        app_label = "canvas_sdk"
        db_table = "canvas_sdk_data_api_patient_001"

    id = models.CharField(max_length=32, db_column="key")
    dbid = models.BigIntegerField(db_column="dbid", primary_key=True)
    first_name = models.CharField()
    last_name = models.CharField()
    birth_date = models.DateField()
    sex_at_birth = models.CharField(choices=SexAtBirth.choices)
    created = models.DateTimeField()
    modified = models.DateTimeField()
    prefix = models.CharField()
    suffix = models.CharField()
    middle_name = models.CharField()
    maiden_name = models.CharField()
    nickname = models.CharField()
    sexual_orientation_term = models.CharField()
    sexual_orientation_code = models.CharField()
    gender_identity_term = models.CharField()
    gender_identity_code = models.CharField()
    preferred_pronouns = models.CharField()
    biological_race_codes = ArrayField(models.CharField())
    last_known_timezone = models.CharField()
    mrn = models.CharField()
    active = models.BooleanField()
    deceased = models.BooleanField()
    deceased_datetime = models.DateTimeField()
    deceased_cause = models.TextField()
    deceased_comment = models.TextField()
    other_gender_description = models.CharField()
    social_security_number = models.CharField()
    administrative_note = models.TextField()
    clinical_note = models.TextField()
    mothers_maiden_name = models.CharField()
    multiple_birth_indicator = models.BooleanField()
    birth_order = models.BigIntegerField()
    default_location_id = models.BigIntegerField()
    default_provider_id = models.BigIntegerField()

    @classmethod
    def find(cls, id: str) -> Self:
        """Find a patient by id."""
        return cls._default_manager.get(id=id)

    def __str__(self) -> str:
        return f"{self.first_name} {self.last_name}"

    def age_at(self, time: arrow.Arrow) -> float:
        """Given a datetime, returns what the patient's age would be at that datetime."""
        age = float(0)
        birth_date = arrow.get(self.birth_date)
        if birth_date.date() < time.date():
            age = time.datetime.year - birth_date.datetime.year
            if time.datetime.month < birth_date.datetime.month or (
                time.datetime.month == birth_date.datetime.month
                and time.datetime.day < birth_date.datetime.day
            ):
                age -= 1

            current_year = birth_date.shift(years=age)
            next_year = birth_date.shift(years=age + 1)
            age += (time.date() - current_year.date()) / (next_year.date() - current_year.date())
        return age



#!/usr/bin/env python

import os
import pickle
from pathlib import Path
from subprocess import STDOUT, CalledProcessError, check_output

import redis

APP_NAME = os.getenv("APP_NAME")

CUSTOMER_IDENTIFIER = os.getenv("CUSTOMER_IDENTIFIER")
PLUGINS_PUBSUB_CHANNEL = os.getenv("PLUGINS_PUBSUB_CHANNEL", default="plugins")

CHANNEL_NAME = f"{CUSTOMER_IDENTIFIER}:{PLUGINS_PUBSUB_CHANNEL}"

REDIS_ENDPOINT = os.getenv("REDIS_ENDPOINT", f"redis://{APP_NAME}-redis:6379")

try:
    CLIENT_ID = Path("/app/container-unique-id.txt").read_text()
except FileNotFoundError:
    CLIENT_ID = "non-unique"


def get_client() -> tuple[redis.Redis, redis.client.PubSub]:
    """Return a Redis client and pubsub object."""
    client = redis.Redis.from_url(REDIS_ENDPOINT)
    pubsub = client.pubsub()

    return client, pubsub


def publish_message(message: dict) -> None:
    """Publish a message to the pubsub channel."""
    client, _ = get_client()

    message_with_id = {**message, "client_id": CLIENT_ID}

    client.publish(CHANNEL_NAME, pickle.dumps(message_with_id))
    client.close()


def main() -> None:
    """Listen for messages on the pubsub channel and restart the plugin-runner."""
    print("plugin-synchronizer: starting")

    _, pubsub = get_client()

    pubsub.psubscribe(CHANNEL_NAME)

    for message in pubsub.listen():
        if not message:
            continue

        message_type = message.get("type", "")

        if message_type != "pmessage":
            continue

        data = pickle.loads(message.get("data", pickle.dumps({})))

        if "action" not in data or "client_id" not in data:
            return

        # Don't respond to our own messages
        if data["client_id"] == CLIENT_ID:
            return

        if data["action"] == "restart":
            # Run the plugin installer process
            try:
                print("plugin-synchronizer: installing plugins")
                check_output(["./manage.py", "install_plugins_v2"], cwd="/app", stderr=STDOUT)
            except CalledProcessError as e:
                print("plugin-synchronizer: `./manage.py install_plugins_v2` failed:", e)

            try:
                print("plugin-synchronizer: sending SIGHUP to plugin-runner")
                check_output(
                    ["circusctl", "signal", "plugin-runner", "1"], cwd="/app", stderr=STDOUT
                )
            except CalledProcessError as e:
                print("plugin-synchronizer: `circusctl signal plugin-runner 1` failed:", e)


if __name__ == "__main__":
    main()



import os
import sys

from dotenv import load_dotenv
from env_tools import env_to_bool

from canvas_sdk.utils.db import get_database_dict_from_url

load_dotenv()

ENV = os.getenv("ENV", "development")
IS_PRODUCTION = ENV == "production"
IS_TESTING = env_to_bool("IS_TESTING", "pytest" in sys.argv[0] or sys.argv[0] == "-c")


INTEGRATION_TEST_URL = os.getenv("INTEGRATION_TEST_URL")
INTEGRATION_TEST_CLIENT_ID = os.getenv("INTEGRATION_TEST_CLIENT_ID")
INTEGRATION_TEST_CLIENT_SECRET = os.getenv("INTEGRATION_TEST_CLIENT_SECRET")

GRAPHQL_ENDPOINT = os.getenv("GRAPHQL_ENDPOINT", "http://localhost:8000/plugins-graphql")

INSTALLED_APPS = ["canvas_sdk"]

SECRET_KEY = os.getenv(
    "SECRET_KEY",
    "This value is the key to securing signed data – it is vital you keep this secure, or attackers could use it to generate their own signed values.",
)

# Use BigAutoField for Default Primary Key field type
DEFAULT_AUTO_FIELD = "django.db.models.BigAutoField"

CANVAS_SDK_DB_NAME = os.getenv("CANVAS_SDK_DB_NAME", "home-app")
CANVAS_SDK_DB_USERNAME = os.getenv("CANVAS_SDK_DB_USERNAME", "app")
CANVAS_SDK_DB_PASSWORD = os.getenv("CANVAS_SDK_DB_PASSWORD", "app")
CANVAS_SDK_DB_HOST = os.getenv("CANVAS_SDK_DB_HOST", "home-app-db")
CANVAS_SDK_DB_PORT = os.getenv("CANVAS_SDK_DB_PORT", "5432")

if os.getenv("DATABASE_URL"):
    database_dict = get_database_dict_from_url()
else:
    database_dict = {
        "ENGINE": "django.db.backends.postgresql",
        "NAME": CANVAS_SDK_DB_NAME,
        "USER": CANVAS_SDK_DB_USERNAME,
        "PASSWORD": CANVAS_SDK_DB_PASSWORD,
        "HOST": CANVAS_SDK_DB_HOST,
        "PORT": CANVAS_SDK_DB_PORT,
    }

DATABASES = {"default": database_dict}


PLUGIN_RUNNER_SIGNING_KEY = os.getenv("PLUGIN_RUNNER_SIGNING_KEY", "")

PLUGIN_DIRECTORY = os.getenv(
    "PLUGIN_DIRECTORY",
    (
        "/plugin-runner/custom-plugins"
        if IS_PRODUCTION
        else "./plugin_runner/tests/data/plugins"
        if IS_TESTING
        else "./custom-plugins"
    ),
)

MANIFEST_FILE_NAME = "CANVAS_MANIFEST.json"

SECRETS_FILE_NAME = "SECRETS.json"



import asyncio
import json
import os
import pathlib
import pkgutil
import signal
import sys
import time
import traceback
from collections import defaultdict
from collections.abc import AsyncGenerator
from types import FrameType
from typing import Any, TypedDict

import grpc
import statsd

from canvas_generated.messages.effects_pb2 import EffectType
from canvas_generated.messages.plugins_pb2 import (
    ReloadPluginsRequest,
    ReloadPluginsResponse,
)
from canvas_generated.services.plugin_runner_pb2_grpc import (
    PluginRunnerServicer,
    add_PluginRunnerServicer_to_server,
)
from canvas_sdk.effects import Effect
from canvas_sdk.events import Event, EventRequest, EventResponse, EventType
from canvas_sdk.protocols import ClinicalQualityMeasure
from canvas_sdk.utils.stats import get_duration_ms, tags_to_line_protocol
from logger import log
from plugin_runner.authentication import token_for_plugin
from plugin_runner.plugin_synchronizer import publish_message
from plugin_runner.sandbox import Sandbox
from settings import MANIFEST_FILE_NAME, PLUGIN_DIRECTORY, SECRETS_FILE_NAME

# when we import plugins we'll use the module name directly so we need to add the plugin
# directory to the path
sys.path.append(PLUGIN_DIRECTORY)

# a global dictionary of loaded plugins
# TODO: create typings here for the subkeys
LOADED_PLUGINS: dict = {}

# a global dictionary of events to handler class names
EVENT_HANDLER_MAP: dict[str, list] = defaultdict(list)


class DataAccess(TypedDict):
    """DataAccess."""

    event: str
    read: list[str]
    write: list[str]


Protocol = TypedDict(
    "Protocol",
    {
        "class": str,
        "data_access": DataAccess,
    },
)


ApplicationConfig = TypedDict(
    "ApplicationConfig",
    {
        "class": str,
        "description": str,
        "icon": str,
        "scope": str,
    },
)


class Components(TypedDict):
    """Components."""

    protocols: list[Protocol]
    commands: list[dict]
    content: list[dict]
    effects: list[dict]
    views: list[dict]
    applications: list[ApplicationConfig]


class PluginManifest(TypedDict):
    """PluginManifest."""

    sdk_version: str
    plugin_version: str
    name: str
    description: str
    components: Components
    secrets: list[dict]
    tags: dict[str, str]
    references: list[str]
    license: str
    diagram: bool
    readme: str


class PluginRunner(PluginRunnerServicer):
    """This process runs provided plugins that register interest in incoming events."""

    def __init__(self) -> None:
        self.statsd_client = statsd.StatsClient()
        super().__init__()

    sandbox: Sandbox

    async def HandleEvent(
        self, request: EventRequest, context: Any
    ) -> AsyncGenerator[EventResponse, None]:
        """This is invoked when an event comes in."""
        event_start_time = time.time()
        event = Event(request)
        event_type = event.type
        event_name = event.name
        relevant_plugins = EVENT_HANDLER_MAP[event_name]

        if event_type in [EventType.PLUGIN_CREATED, EventType.PLUGIN_UPDATED]:
            plugin_name = event.target.id
            # filter only for the plugin(s) that were created/updated
            relevant_plugins = [p for p in relevant_plugins if p.startswith(f"{plugin_name}:")]

        effect_list = []

        for plugin_name in relevant_plugins:
            plugin = LOADED_PLUGINS[plugin_name]
            handler_class = plugin["class"]
            base_plugin_name = plugin_name.split(":")[0]

            secrets = plugin.get("secrets", {})
            secrets["graphql_jwt"] = token_for_plugin(plugin_name=plugin_name, audience="home")

            try:
                handler = handler_class(event, secrets)
                classname = (
                    handler.__class__.__name__
                    if isinstance(handler, ClinicalQualityMeasure)
                    else None
                )

                compute_start_time = time.time()
                _effects = await asyncio.get_running_loop().run_in_executor(None, handler.compute)
                effects = [
                    Effect(
                        type=effect.type,
                        payload=effect.payload,
                        plugin_name=base_plugin_name,
                        classname=classname,
                    )
                    for effect in _effects
                ]

                effects = validate_effects(effects)

                apply_effects_to_context(effects, event=event)

                compute_duration = get_duration_ms(compute_start_time)

                log.info(f"{plugin_name}.compute() completed ({compute_duration} ms)")
                statsd_tags = tags_to_line_protocol({"plugin": plugin_name})
                self.statsd_client.timing(
                    f"plugins.protocol_duration_ms,{statsd_tags}",
                    delta=compute_duration,
                )
            except Exception as e:
                for error_line_with_newlines in traceback.format_exception(e):
                    for error_line in error_line_with_newlines.split("\n"):
                        log.error(error_line)
                continue

            effect_list += effects

        event_duration = get_duration_ms(event_start_time)

        # Don't log anything if a plugin handler didn't actually run.
        if relevant_plugins:
            log.info(f"Responded to Event {event_name} ({event_duration} ms)")
            statsd_tags = tags_to_line_protocol({"event": event_name})
            self.statsd_client.timing(
                f"plugins.event_duration_ms,{statsd_tags}",
                delta=event_duration,
            )

        yield EventResponse(success=True, effects=effect_list)

    async def ReloadPlugins(
        self, request: ReloadPluginsRequest, context: Any
    ) -> AsyncGenerator[ReloadPluginsResponse, None]:
        """This is invoked when we need to reload plugins."""
        try:
            load_plugins()
            publish_message({"action": "restart"})
        except ImportError:
            yield ReloadPluginsResponse(success=False)
        else:
            yield ReloadPluginsResponse(success=True)


def validate_effects(effects: list[Effect]) -> list[Effect]:
    """Validates the effects based on predefined rules.
    Keeps only the first AUTOCOMPLETE_SEARCH_RESULTS effect and preserve all non-search-related effects.
    """
    seen_autocomplete = False
    validated_effects = []

    for effect in effects:
        if effect.type == EffectType.AUTOCOMPLETE_SEARCH_RESULTS:
            if seen_autocomplete:
                log.warning("Discarding additional AUTOCOMPLETE_SEARCH_RESULTS effect.")
                continue
            seen_autocomplete = True
        validated_effects.append(effect)

    return validated_effects


def apply_effects_to_context(effects: list[Effect], event: Event) -> Event:
    """Applies AUTOCOMPLETE_SEARCH_RESULTS effects to the event context.
    If we are dealing with a search event, we need to update the context with the search results.
    """
    event_name = event.name

    # Skip if the event is not a search event
    if not event_name.endswith("__PRE_SEARCH") and not event_name.endswith("__POST_SEARCH"):
        return event

    for effect in effects:
        if effect.type == EffectType.AUTOCOMPLETE_SEARCH_RESULTS:
            event.context["results"] = json.loads(effect.payload)
            # Stop processing effects if we've found a AUTOCOMPLETE_SEARCH_RESULTS
            break

    return event


def handle_hup_cb(_signum: int, _frame: FrameType | None) -> None:
    """handle_hup_cb."""
    log.info("Received SIGHUP, reloading plugins...")
    load_plugins()


def find_modules(base_path: pathlib.Path, prefix: str | None = None) -> list[str]:
    """Find all modules in the specified package path."""
    modules: list[str] = []

    for _, module_name, is_pkg in pkgutil.iter_modules(
        [base_path.as_posix()],
    ):
        if is_pkg:
            modules = modules + find_modules(
                base_path / module_name,
                prefix=f"{prefix}.{module_name}" if prefix else module_name,
            )
        else:
            modules.append(f"{prefix}.{module_name}" if prefix else module_name)

    return modules


def sandbox_from_package(package_path: pathlib.Path) -> dict[str, Any]:
    """Sandbox the code execution."""
    package_name = package_path.name
    available_modules = find_modules(package_path)
    sandboxes = {}

    for module_name in available_modules:
        result = sandbox_from_module(package_path, module_name)
        full_module_name = f"{package_name}.{module_name}"
        sandboxes[full_module_name] = result

    return sandboxes


def sandbox_from_module(package_path: pathlib.Path, module_name: str) -> Any:
    """Sandbox the code execution."""
    module_path = package_path / str(module_name.replace(".", "/") + ".py")

    if not module_path.exists():
        raise ModuleNotFoundError(f'Could not load module "{module_name}"')

    source_code = module_path.read_text()

    full_module_name = f"{package_path.name}.{module_name}"

    sandbox = Sandbox(source_code, namespace=full_module_name)
    return sandbox.execute()


def load_or_reload_plugin(path: pathlib.Path) -> None:
    """Given a path, load or reload a plugin."""
    log.info(f"Loading {path}")

    manifest_file = path / MANIFEST_FILE_NAME
    manifest_json_str = manifest_file.read_text()

    # the name is the folder name underneath the plugins directory
    name = path.name

    try:
        manifest_json: PluginManifest = json.loads(manifest_json_str)
    except Exception as e:
        log.error(f'Unable to load plugin "{name}": {e}')
        return

    secrets_file = path / SECRETS_FILE_NAME

    secrets_json = {}
    if secrets_file.exists():
        try:
            secrets_json = json.load(secrets_file.open())
        except Exception as e:
            log.error(f'Unable to load secrets for plugin "{name}": {str(e)}')

    # TODO add existing schema validation from Michela here
    try:
        handlers = manifest_json["components"].get("protocols", []) + manifest_json[
            "components"
        ].get("applications", [])
        results = sandbox_from_package(path)
    except Exception as e:
        log.error(f'Unable to load plugin "{name}": {str(e)}')
        return

    for handler in handlers:
        # TODO add class colon validation to existing schema validation
        # TODO when we encounter an exception here, disable the plugin in response
        try:
            handler_module, handler_class = handler["class"].split(":")
            name_and_class = f"{name}:{handler_module}:{handler_class}"
        except ValueError:
            log.error(f"Unable to parse class for plugin '{name}': '{handler['class']}'")
            continue

        try:
            if name_and_class in LOADED_PLUGINS:
                log.info(f"Reloading plugin '{name_and_class}'")

                result = results[handler_module]

                LOADED_PLUGINS[name_and_class]["active"] = True

                LOADED_PLUGINS[name_and_class]["class"] = result[handler_class]
                LOADED_PLUGINS[name_and_class]["sandbox"] = result
                LOADED_PLUGINS[name_and_class]["secrets"] = secrets_json
            else:
                log.info(f"Loading plugin '{name_and_class}'")

                result = results[handler_module]

                LOADED_PLUGINS[name_and_class] = {
                    "active": True,
                    "class": result[handler_class],
                    "sandbox": result,
                    "handler": handler,
                    "secrets": secrets_json,
                }
        except Exception as err:
            log.error(f"Error importing module '{name_and_class}': {err}")
            for error_line in traceback.format_exception(err):
                log.error(error_line)


def refresh_event_type_map() -> None:
    """Ensure the event subscriptions are up to date."""
    EVENT_HANDLER_MAP.clear()

    for name, plugin in LOADED_PLUGINS.items():
        if hasattr(plugin["class"], "RESPONDS_TO"):
            responds_to = plugin["class"].RESPONDS_TO

            if isinstance(responds_to, str):
                EVENT_HANDLER_MAP[responds_to].append(name)
            elif isinstance(responds_to, list):
                for event in responds_to:
                    EVENT_HANDLER_MAP[event].append(name)
            else:
                log.warning(f"Unknown RESPONDS_TO type: {type(responds_to)}")


def load_plugins(specified_plugin_paths: list[str] | None = None) -> None:
    """Load the plugins."""
    # first mark each plugin as inactive since we want to remove it from
    # LOADED_PLUGINS if it no longer exists on disk
    for plugin in LOADED_PLUGINS.values():
        plugin["active"] = False

    if specified_plugin_paths is not None:
        # convert to Paths
        plugin_paths = [pathlib.Path(name) for name in specified_plugin_paths]

        for plugin_path in plugin_paths:
            # when we import plugins we'll use the module name directly so we need to add the plugin
            # directory to the path
            path_to_append = pathlib.Path(".") / plugin_path.parent
            sys.path.append(path_to_append.as_posix())
    else:
        candidates = os.listdir(PLUGIN_DIRECTORY)

        # convert to Paths
        plugin_paths = [pathlib.Path(os.path.join(PLUGIN_DIRECTORY, name)) for name in candidates]

    # get all directories under the plugin directory
    plugin_paths = [path for path in plugin_paths if path.is_dir()]

    # filter to only the directories containing a manifest file
    plugin_paths = [path for path in plugin_paths if (path / MANIFEST_FILE_NAME).exists()]

    # load or reload each plugin
    for plugin_path in plugin_paths:
        load_or_reload_plugin(plugin_path)

    # if a plugin has been uninstalled/disabled remove it from LOADED_PLUGINS
    for name, plugin in LOADED_PLUGINS.copy().items():
        if not plugin["active"]:
            del LOADED_PLUGINS[name]

    refresh_event_type_map()


_cleanup_coroutines = []


async def serve(specified_plugin_paths: list[str] | None = None) -> None:
    """Run the server."""
    port = "50051"

    server = grpc.aio.server()
    server.add_insecure_port("127.0.0.1:" + port)

    add_PluginRunnerServicer_to_server(PluginRunner(), server)

    log.info(f"Starting server, listening on port {port}")

    load_plugins(specified_plugin_paths)

    await server.start()

    async def server_graceful_shutdown() -> None:
        log.info("Starting graceful shutdown...")
        await server.stop(5)

    _cleanup_coroutines.append(server_graceful_shutdown())

    await server.wait_for_termination()


def run_server(specified_plugin_paths: list[str] | None = None) -> None:
    """Run the server."""
    loop = asyncio.new_event_loop()

    asyncio.set_event_loop(loop)

    signal.signal(signal.SIGHUP, handle_hup_cb)

    try:
        loop.run_until_complete(serve(specified_plugin_paths))
    except KeyboardInterrupt:
        pass
    finally:
        loop.run_until_complete(*_cleanup_coroutines)
        loop.close()


if __name__ == "__main__":
    run_server()



from canvas_sdk.commands.base import _BaseCommand as BaseCommand


class InstructCommand(BaseCommand):
    """A class for managing an Instruct command within a specific note."""

    class Meta:
        key = "instruct"
        commit_required_fields = ("instruction",)

    instruction: str | None = None
    comment: str | None = None

    @property
    def values(self) -> dict:
        """The Instruct command's field values."""
        return {"instruction": self.instruction, "comment": self.comment}



from canvas_sdk.commands.base import _BaseCommand as BaseCommand


class FamilyHistoryCommand(BaseCommand):
    """A class for managing a Family History command within a specific note."""

    class Meta:
        key = "familyHistory"
        commit_required_fields = ("family_history",)

    family_history: str | None = None
    relative: str | None = None
    note: str | None = None

    @property
    def values(self) -> dict:
        """The Family History command's field values."""
        return {"family_history": self.family_history, "relative": self.relative, "note": self.note}



from canvas_sdk.commands.base import _BaseCommand as BaseCommand


class LabOrderCommand(BaseCommand):
    """A class for managing a Lab Order command within a specific note."""

    class Meta:
        key = "labOrder"
        commit_required_fields = (
            "lab_partner",
            "tests_order_codes",
            "ordering_provider",
            "diagnosis_codes",
        )

    lab_partner: str | None = None
    tests_order_codes: list[str] = []
    ordering_provider_key: str | None = None
    diagnosis_codes: list[str] = []
    fasting_required: bool = False
    comment: str | None = None

    @property
    def values(self) -> dict:
        """The Lab Order command's field values."""
        return {
            "lab_partner": self.lab_partner,
            "tests_order_codes": self.tests_order_codes,
            "ordering_provider_key": self.ordering_provider_key,
            "diagnosis_codes": self.diagnosis_codes,
            "fasting_required": self.fasting_required,
            "comment": self.comment,
        }



import pytest

from canvas_sdk.effects import Effect
from canvas_sdk.effects.launch_modal import LaunchModalEffect
from canvas_sdk.events import Event, EventRequest, EventType
from canvas_sdk.handlers.application import Application


class TestApplication(Application):
    """A concrete implementation of the Application class for testing."""

    def on_open(self) -> Effect:
        """Handle the application open event by returning a mock effect."""
        return LaunchModalEffect(url="https://example.com").apply()


@pytest.fixture
def app_instance(event: Event) -> TestApplication:
    """Provide an instance of the TestApplication with a mocked event."""
    app = TestApplication(event)
    return app


def test_compute_event_not_targeted() -> None:
    """Test that compute filters out events not targeted for the app."""
    request = EventRequest(type=EventType.APPLICATION__ON_OPEN, target="some_identifier")
    event = Event(request)
    app = TestApplication(event)

    result = app.compute()

    assert result == [], "Expected no effects if the event target is not the app identifier"


def test_compute_event_targeted() -> None:
    """Test that compute processes events targeted for the app."""
    request = EventRequest(
        type=EventType.APPLICATION__ON_OPEN,
        target=f"{TestApplication.__module__}:{TestApplication.__qualname__}",
    )
    event = Event(request)
    app = TestApplication(event)
    result = app.compute()

    assert len(result) == 1, "Expected a single effect if the event target is the app identifier"
    assert isinstance(result[0], Effect), "Effect should be an instance of Effect"


def test_identifier_property() -> None:
    """Test the identifier property of the Application class."""
    expected_identifier = f"{TestApplication.__module__}:{TestApplication.__qualname__}"
    request = EventRequest(
        type=EventType.APPLICATION__ON_OPEN,
        target=f"{TestApplication.__module__}:{TestApplication.__qualname__}",
    )
    event = Event(request)
    app = TestApplication(event)

    assert app.identifier == expected_identifier, "The identifier property is incorrect"


def test_abstract_method_on_open() -> None:
    """Test that the abstract method on_open must be implemented."""
    with pytest.raises(TypeError):
        Application(Event(EventRequest(type=EventType.UNKNOWN)))  # type: ignore[abstract]



import json
import os
from collections.abc import Generator
from pathlib import Path

import pytest
import typer

from canvas_cli.utils.context import CLIContext


class CLIContextTestHelper(CLIContext):
    """CLIContext subclass that defines some properties we can test."""

    _persistent_mock_property: str | None = None
    _transient_mock_property: bool = False

    @property
    def persistent_mock_property(self) -> str | None:
        """Mock persistent property."""
        return self._persistent_mock_property

    @persistent_mock_property.setter
    @CLIContext.persistent
    def persistent_mock_property(self, new_persistent_mock_property: str | None) -> None:
        self._persistent_mock_property = new_persistent_mock_property

    @property
    def transient_mock_property(self) -> bool:
        """Mock transient property."""
        return self._transient_mock_property

    @transient_mock_property.setter
    def transient_mock_property(self, new_transient_mock_property: bool) -> None:
        self._transient_mock_property = new_transient_mock_property


@pytest.fixture
def config_file(tmp_path: Path) -> Generator[Path, None, None]:
    """Fixture that yields an empty config file and cleans up after itself."""
    config_file = tmp_path / "mock_config.json"
    yield config_file
    os.remove(config_file)


@pytest.fixture
def valid_mock_config_file(config_file: Path) -> Path:
    """Fixture that yields a valid mock config file."""
    mock_config = {
        "persistent_mock_property": "mock-value",
    }
    with open(config_file, "w") as f:
        json.dump(mock_config, f)

    return config_file


@pytest.fixture
def invalid_json_mock_config_file(config_file: Path) -> Path:
    """Fixture that yields an invalid config file."""
    config_file.write_text("Absolutely invalid JSON")

    return config_file


@pytest.fixture
def invalid_properties_mock_config_file(config_file: Path) -> Path:
    """Fixture that yields a valid mock config file."""
    mock_config = {
        "unknown-property": "mock-value",
    }
    with open(config_file, "w") as f:
        json.dump(mock_config, f)

    return config_file


def test_valid_load_from_file(valid_mock_config_file: Path) -> None:
    """Test loading a valid config file."""
    context = CLIContextTestHelper()
    context.load_from_file(valid_mock_config_file)

    assert context.persistent_mock_property == "mock-value"


def test_invalid_load_from_file(invalid_json_mock_config_file: Path) -> None:
    """Test loading an invalid config file aborts the execution."""
    context = CLIContextTestHelper()

    with pytest.raises(typer.Abort):
        context.load_from_file(invalid_json_mock_config_file)


def test_load_invalid_property(invalid_properties_mock_config_file: Path) -> None:
    """Since we dynamically load the properties, this test ensures that unknown properties don't throw exceptions."""
    context = CLIContextTestHelper()

    context.load_from_file(invalid_properties_mock_config_file)


def test_config_persistence(valid_mock_config_file: Path) -> None:
    """Test marking a property with @persistent stores the value in the config file."""
    context = CLIContextTestHelper()
    context.load_from_file(valid_mock_config_file)

    assert context.persistent_mock_property == "mock-value"

    context.persistent_mock_property = "new-value"

    # This won't ever happen but since the values are in memory, we need to create a new context instance,
    # as if mimicking a new program launch
    context_b = CLIContextTestHelper()
    context_b.load_from_file(valid_mock_config_file)

    assert context_b.persistent_mock_property == "new-value"


def test_config_transience(valid_mock_config_file: Path) -> None:
    """Test the properties transient default."""
    context = CLIContextTestHelper()
    context.load_from_file(valid_mock_config_file)

    assert context.transient_mock_property is False
    context.transient_mock_property = True

    # This won't ever happen but since the values are in memory, we need to create a new context instance,
    # as if mimicking a new program launch
    context_b = CLIContextTestHelper()
    context_b.load_from_file(valid_mock_config_file)

    assert context_b.transient_mock_property is False



import dataclasses
import json
from functools import cached_property
from typing import Any

from django.apps import apps
from django.db import models

from canvas_generated.messages.events_pb2 import Event as EventRequest
from canvas_generated.messages.events_pb2 import EventType


@dataclasses.dataclass
class TargetType:
    """The target of the event."""

    id: str
    type: type[models.Model] | None

    @cached_property
    def instance(self) -> models.Model | None:
        """Return the instance of the target."""
        return self.type._default_manager.filter(id=self.id).first() if self.type else None


class Event:
    """An event that occurs in the Canvas environment."""

    name: str
    type: EventType
    context: dict[str, Any]
    target: TargetType

    def __init__(self, event_request: EventRequest) -> None:
        try:
            target_model = apps.get_model(
                app_label="canvas_sdk", model_name=event_request.target_type
            )
        except LookupError:
            target_model = None

        try:
            context = json.loads(event_request.context)
        except ValueError:
            context = {}

        self.type = event_request.type
        self.name = EventType.Name(self.type)
        self.context = context
        self.target = TargetType(id=event_request.target, type=target_model)



import decimal
from datetime import date, datetime
from typing import get_origin

import pytest
import requests

import settings
from canvas_sdk.commands.base import _BaseCommand
from canvas_sdk.commands.constants import ClinicalQuantity, Coding
from canvas_sdk.commands.tests.test_utils import (
    COMMANDS,
    MaskedValue,
    create_new_note,
    get_token,
)


@pytest.fixture(scope="session")
def token() -> MaskedValue:
    """Get a valid token."""
    return get_token()


@pytest.fixture(scope="session")
def new_note(token: MaskedValue) -> dict:
    """Create a new note."""
    return create_new_note(token)


@pytest.fixture
def command_type_map() -> dict[str, type]:
    """Map of command field types to their corresponding Python types."""
    return {
        "AutocompleteField": str,
        "MultiLineTextField": str,
        "TextField": str,
        "ChoiceField": str,
        "DateField": datetime,
        "ApproximateDateField": date,
        "IntegerField": int,
        "DecimalField": decimal.Decimal,
    }


@pytest.mark.integtest
@pytest.mark.parametrize(
    "Command",
    COMMANDS,
)
def test_command_schema_matches_command_api(
    token: MaskedValue,
    command_type_map: dict[str, str],
    new_note: dict,
    Command: _BaseCommand,
) -> None:
    """Test that the command schema matches the command API."""
    # first create the command in the new note
    data = {"noteKey": new_note["externallyExposableId"], "schemaKey": Command.Meta.key}
    headers = {"Authorization": f"Bearer {token.value}"}
    url = f"{settings.INTEGRATION_TEST_URL}/core/api/v1/commands/"
    command_resp = requests.post(url, headers=headers, data=data).json()
    assert "uuid" in command_resp
    command_uuid = command_resp["uuid"]

    # next, request the fields of the newly created command
    url = f"{settings.INTEGRATION_TEST_URL}/core/api/v1/commands/{command_uuid}/fields/"
    command_fields_resp = requests.get(url, headers=headers).json()
    assert command_fields_resp["schema"] == Command.Meta.key

    command_fields = command_fields_resp["fields"]
    if Command.Meta.key == "questionnaire":
        # questionnaire's fields vary per questionnaire, so just check the first two fields which never vary
        command_fields = command_fields[:2]
    expected_fields = Command.command_schema()
    assert len(command_fields) == len(expected_fields)

    for actual_field in command_fields:
        name = actual_field["name"]
        assert name in expected_fields
        expected_field = expected_fields[name]

        assert expected_field["required"] == actual_field["required"]

        expected_type = expected_field["type"]
        if expected_type is Coding:
            expected_type = expected_type.__annotations__["code"]

        if expected_type is ClinicalQuantity:
            expected_type = expected_type.__annotations__["representative_ndc"]

        actual_type = command_type_map.get(actual_field["type"])
        if actual_field["type"] == "AutocompleteField" and name[-1] == "s":
            # this condition initially created for Prescribe.indications,
            # but could apply to other AutocompleteField fields that are lists
            # making the assumption here that if the field ends in 's' (like indications), it is a list
            assert get_origin(expected_type) is list

        else:
            assert expected_type == actual_type

        if (choices := actual_field["choices"]) is None:
            assert expected_field["choices"] is None
            continue

        assert len(expected_field["choices"]) == len(choices)
        for choice in choices:
            assert choice["value"] in expected_field["choices"]



from datetime import date

from pydantic import Field

from canvas_sdk.commands.base import _BaseCommand as BaseCommand


class MedicalHistoryCommand(BaseCommand):
    """A class for managing a Medical History command within a specific note."""

    class Meta:
        key = "medicalHistory"
        commit_required_fields = ("past_medical_history",)

    past_medical_history: str | None = None
    approximate_start_date: date | None = None
    approximate_end_date: date | None = None
    show_on_condition_list: bool = True
    comments: str | None = Field(max_length=1000, default=None)

    @property
    def values(self) -> dict:
        """The Medical History command's field values."""
        return {
            "past_medical_history": self.past_medical_history,
            "approximate_start_date": (
                self.approximate_start_date.isoformat() if self.approximate_start_date else None
            ),
            "approximate_end_date": (
                self.approximate_end_date.isoformat() if self.approximate_end_date else None
            ),
            "show_on_condition_list": self.show_on_condition_list,
            "comments": self.comments,
        }



from canvas_sdk.effects.banner_alert.add_banner_alert import AddBannerAlert
from canvas_sdk.effects.banner_alert.remove_banner_alert import RemoveBannerAlert

__all__ = ("AddBannerAlert", "RemoveBannerAlert")



from typing import Any

from pydantic import Field

from canvas_sdk.effects.base import EffectType, _BaseEffect


class IntakeFormResults(_BaseEffect):
    """An Effect that will decide which intake forms (questionnaires) appear on the patient portal."""

    class Meta:
        effect_type = EffectType.PATIENT_PORTAL__INTAKE_FORM_RESULTS

    questionnaire_ids: list = Field(min_length=0)

    @property
    def values(self) -> dict[str, Any]:
        """The IntakeFormResults's values."""
        return {"questionnaire_ids": [str(q) for q in self.questionnaire_ids]}

    @property
    def effect_payload(self) -> dict[str, Any]:
        """The payload of the effect."""
        return {"data": self.values}



from decimal import Decimal
from enum import Enum

from pydantic import Field, conlist

from canvas_sdk.commands.base import _BaseCommand
from canvas_sdk.commands.constants import ClinicalQuantity


class PrescribeCommand(_BaseCommand):
    """A class for managing a Prescribe command within a specific note."""

    class Meta:
        key = "prescribe"
        commit_required_fields = (
            "fdb_code",
            "sig",
            "quantity_to_dispense",
            "type_to_dispense",
            "refills",
            "substitutions",
            "prescriber_id",
        )

    class Substitutions(Enum):
        ALLOWED = "allowed"
        NOT_ALLOWED = "not_allowed"

    fdb_code: str | None = Field(default=None, json_schema_extra={"commands_api_name": "prescribe"})
    icd10_codes: conlist(str, max_length=2) = Field(  # type: ignore[valid-type]
        [], json_schema_extra={"commands_api_name": "indications"}
    )
    sig: str = ""
    days_supply: int | None = None
    quantity_to_dispense: Decimal | float | int | None = None
    type_to_dispense: ClinicalQuantity | None = None
    refills: int | None = None
    substitutions: Substitutions | None = None
    pharmacy: str | None = None
    prescriber_id: str | None = Field(
        default=None, json_schema_extra={"commands_api_name": "prescriber"}
    )
    note_to_pharmacist: str | None = None

    @property
    def values(self) -> dict:
        """The Prescribe command's field values."""
        return {
            "fdb_code": self.fdb_code,
            "icd10_codes": self.icd10_codes,
            "sig": self.sig,
            "days_supply": self.days_supply,
            "quantity_to_dispense": (
                str(Decimal(self.quantity_to_dispense)) if self.quantity_to_dispense else None
            ),
            "type_to_dispense": self.type_to_dispense if self.type_to_dispense else None,
            "refills": self.refills,
            "substitutions": self.substitutions.value if self.substitutions else None,
            "pharmacy": self.pharmacy,
            "prescriber_id": self.prescriber_id,
            "note_to_pharmacist": self.note_to_pharmacist,
        }



from datetime import datetime
from enum import Enum

from canvas_sdk.commands.base import _BaseCommand


class GoalCommand(_BaseCommand):
    """A class for managing a Goal command within a specific note."""

    class Meta:
        key = "goal"
        commit_required_fields = ("goal_statement", "start_date")

    class Priority(Enum):
        HIGH = "high-priority"
        MEDIUM = "medium-priority"
        LOW = "low-priority"

    class AchievementStatus(Enum):
        IN_PROGRESS = "in-progress"
        IMPROVING = "improving"
        WORSENING = "worsening"
        NO_CHANGE = "no-change"
        ACHIEVED = "achieved"
        SUSTAINING = "sustaining"
        NOT_ACHIEVED = "not-achieved"
        NO_PROGRESS = "no-progress"
        NOT_ATTAINABLE = "not-attainable"

    goal_statement: str = ""
    start_date: datetime = datetime.now()
    due_date: datetime | None = None
    achievement_status: AchievementStatus | None = None
    priority: Priority | None = None
    progress: str | None = None

    @property
    def values(self) -> dict:
        """The Goal command's field values."""
        return {
            "goal_statement": self.goal_statement,
            "start_date": (self.start_date.isoformat() if self.start_date else None),
            "due_date": (self.due_date.isoformat() if self.due_date else None),
            "achievement_status": (
                self.achievement_status.value if self.achievement_status else None
            ),
            "priority": (self.priority.value if self.priority else None),
            "progress": self.progress,
        }



from collections.abc import Generator
from datetime import datetime

import pytest

from canvas_sdk.commands.tests.test_utils import (
    COMMANDS,
    MaskedValue,
    clean_up_files_and_plugins,
    create_new_note,
    get_original_note_body_commands,
    get_token,
    install_plugin,
    trigger_plugin_event,
    write_protocol_code,
)


@pytest.fixture(scope="session")
def token() -> MaskedValue:
    """Get a valid token."""
    return get_token()


@pytest.fixture(scope="session")
def new_note(token: MaskedValue) -> dict:
    """Create a new note."""
    return create_new_note(token)


@pytest.fixture(scope="session")
def plugin_name() -> str:
    """The plugin name to be used."""
    return f"commands{datetime.now().timestamp()}".replace(".", "")


@pytest.fixture(scope="session")
def write_and_install_protocol_and_clean_up(
    plugin_name: str, token: MaskedValue, new_note: dict
) -> Generator[None, None, None]:
    """Write the protocol code, install the plugin, and clean up after the test."""
    write_protocol_code(new_note["externallyExposableId"], plugin_name, COMMANDS)
    install_plugin(plugin_name, token)

    yield

    clean_up_files_and_plugins(plugin_name, token)


@pytest.mark.integtest
def test_protocol_that_inserts_every_command(
    write_and_install_protocol_and_clean_up: None, token: MaskedValue, new_note: dict
) -> None:
    """Test that the protocol inserts every command."""
    trigger_plugin_event(token)

    commands_in_body = get_original_note_body_commands(new_note["id"], token)
    command_keys = [c.Meta.key for c in COMMANDS]

    assert len(command_keys) == len(commands_in_body)
    for i, command_key in enumerate(command_keys):
        assert commands_in_body[i] == command_key



from ...test_module_imports_outside_plugin_v2.other_module.base import (
    import_me as other_module_import_me,
)


def import_me() -> list[str]:
    """Test method."""
    return other_module_import_me()



import shutil
from collections.abc import Generator
from pathlib import Path
from unittest.mock import MagicMock, patch

import pytest

from canvas_generated.messages.effects_pb2 import EffectType
from canvas_generated.messages.plugins_pb2 import ReloadPluginsRequest
from canvas_sdk.events import EventRequest, EventType
from plugin_runner.plugin_runner import (
    EVENT_HANDLER_MAP,
    LOADED_PLUGINS,
    PluginRunner,
    load_plugins,
    sandbox_from_package,
)


@pytest.fixture
def setup_test_plugin(request: pytest.FixtureRequest) -> Generator[Path, None, None]:
    """Copies a specified plugin from the fixtures directory to the data directory
    and removes it after the test.

    Parameters:
    - request.param: The name of the plugin package to copy.

    Yields:
    - Path to the copied plugin directory.
    """
    # Define base directories
    base_dir = Path("./plugin_runner/tests")
    fixture_plugin_dir = base_dir / "fixtures" / "plugins"
    data_plugin_dir = base_dir / "data" / "plugins"

    # The plugin name should be passed as a parameter to the fixture
    plugin_name = request.param  # Expected to be a str
    src_plugin_path = fixture_plugin_dir / plugin_name
    dest_plugin_path = data_plugin_dir / plugin_name

    # Ensure the data plugin directory exists
    data_plugin_dir.mkdir(parents=True, exist_ok=True)

    # Copy the specific plugin from fixtures to data
    try:
        shutil.copytree(src_plugin_path, dest_plugin_path)
        yield dest_plugin_path  # Provide the path to the test
    finally:
        # Cleanup: remove data/plugins directory after the test
        if dest_plugin_path.exists():
            shutil.rmtree(dest_plugin_path)


@pytest.fixture
def plugin_runner() -> PluginRunner:
    """Fixture to initialize PluginRunner with mocks."""
    runner = PluginRunner()
    runner.statsd_client = MagicMock()
    return runner


@pytest.mark.parametrize("setup_test_plugin", ["example_plugin"], indirect=True)
def test_load_plugins_with_valid_plugin(setup_test_plugin: Path) -> None:
    """Test loading plugins with a valid plugin."""
    load_plugins()

    assert "example_plugin:example_plugin.protocols.my_protocol:Protocol" in LOADED_PLUGINS
    assert (
        LOADED_PLUGINS["example_plugin:example_plugin.protocols.my_protocol:Protocol"]["active"]
        is True
    )


@pytest.mark.asyncio
@pytest.mark.parametrize("setup_test_plugin", ["test_module_imports_plugin"], indirect=True)
async def test_load_plugins_with_plugin_that_imports_other_modules_within_plugin_package(
    setup_test_plugin: Path, plugin_runner: PluginRunner
) -> None:
    """Test loading plugins with a valid plugin that imports other modules within the current plugin package."""
    load_plugins()
    assert (
        "test_module_imports_plugin:test_module_imports_plugin.protocols.my_protocol:Protocol"
        in LOADED_PLUGINS
    )
    assert (
        LOADED_PLUGINS[
            "test_module_imports_plugin:test_module_imports_plugin.protocols.my_protocol:Protocol"
        ]["active"]
        is True
    )

    result = [
        response
        async for response in plugin_runner.HandleEvent(EventRequest(type=EventType.UNKNOWN), None)
    ]

    assert len(result) == 1
    assert result[0].success is True
    assert len(result[0].effects) == 1
    assert result[0].effects[0].type == EffectType.LOG
    assert result[0].effects[0].payload == "Successfully imported!"


@pytest.mark.parametrize(
    "setup_test_plugin",
    [
        "test_module_imports_outside_plugin_v1",
        "test_module_imports_outside_plugin_v2",
        "test_module_imports_outside_plugin_v3",
    ],
    indirect=True,
)
def test_load_plugins_with_plugin_that_imports_other_modules_outside_plugin_package(
    setup_test_plugin: Path,
) -> None:
    """Test loading plugins with an invalid plugin that imports other modules outside the current plugin package."""
    with pytest.raises(ImportError, match="is not an allowed import"):
        sandbox_from_package(setup_test_plugin)


@pytest.mark.parametrize("setup_test_plugin", ["example_plugin"], indirect=True)
def test_reload_plugin(setup_test_plugin: Path) -> None:
    """Test reloading a plugin."""
    load_plugins()
    load_plugins()

    assert "example_plugin:example_plugin.protocols.my_protocol:Protocol" in LOADED_PLUGINS
    assert (
        LOADED_PLUGINS["example_plugin:example_plugin.protocols.my_protocol:Protocol"]["active"]
        is True
    )


@pytest.mark.parametrize("setup_test_plugin", ["example_plugin"], indirect=True)
def test_remove_plugin_should_be_removed_from_loaded_plugins(setup_test_plugin: Path) -> None:
    """Test removing a plugin."""
    load_plugins()
    assert "example_plugin:example_plugin.protocols.my_protocol:Protocol" in LOADED_PLUGINS
    shutil.rmtree(setup_test_plugin)
    load_plugins()
    assert "example_plugin:example_plugin.protocols.my_protocol:Protocol" not in LOADED_PLUGINS


@pytest.mark.parametrize("setup_test_plugin", ["example_plugin"], indirect=True)
def test_load_plugins_should_refresh_event_protocol_map(setup_test_plugin: Path) -> None:
    """Test that the event protocol map is refreshed when loading plugins."""
    assert EVENT_HANDLER_MAP == {}
    load_plugins()
    assert EventType.Name(EventType.UNKNOWN) in EVENT_HANDLER_MAP
    assert EVENT_HANDLER_MAP[EventType.Name(EventType.UNKNOWN)] == [
        "example_plugin:example_plugin.protocols.my_protocol:Protocol"
    ]


@pytest.mark.asyncio
@pytest.mark.parametrize("setup_test_plugin", ["example_plugin"], indirect=True)
async def test_handle_plugin_event_returns_expected_result(
    setup_test_plugin: Path, plugin_runner: PluginRunner
) -> None:
    """Test that HandleEvent successfully calls the relevant plugins and returns the expected result."""
    load_plugins()

    event = EventRequest(type=EventType.UNKNOWN)

    result = []
    async for response in plugin_runner.HandleEvent(event, None):
        result.append(response)

    assert len(result) == 1
    assert result[0].success is True
    assert len(result[0].effects) == 1
    assert result[0].effects[0].type == EffectType.LOG
    assert result[0].effects[0].payload == "Hello, world!"


@pytest.mark.asyncio
@pytest.mark.parametrize("setup_test_plugin", ["example_plugin"], indirect=True)
async def test_reload_plugins_event_handler_successfully_loads_plugins(
    setup_test_plugin: Path, plugin_runner: PluginRunner
) -> None:
    """Test ReloadPlugins Event handler successfully loads plugins."""
    with patch("plugin_runner.plugin_runner.publish_message", MagicMock()) as mock_publish_message:
        request = ReloadPluginsRequest()

        result = []
        async for response in plugin_runner.ReloadPlugins(request, None):
            result.append(response)

        mock_publish_message.assert_called_once_with({"action": "restart"})

    assert len(result) == 1
    assert result[0].success is True
    assert "example_plugin:example_plugin.protocols.my_protocol:Protocol" in LOADED_PLUGINS


@pytest.mark.asyncio
async def test_reload_plugins_import_error(plugin_runner: PluginRunner) -> None:
    """Test ReloadPlugins response when an ImportError occurs."""
    request = ReloadPluginsRequest()

    with patch("plugin_runner.plugin_runner.load_plugins", side_effect=ImportError):
        responses = []
        async for response in plugin_runner.ReloadPlugins(request, None):
            responses.append(response)

        assert len(responses) == 1
        assert responses[0].success is False



from typing import Any

from canvas_sdk.effects.base import EffectType, _BaseEffect


class RemoveBannerAlert(_BaseEffect):
    """
    An Effect that will remove/inactivate a banner alert in Canvas.
    """

    class Meta:
        effect_type = EffectType.REMOVE_BANNER_ALERT
        apply_required_fields = ("patient_id|patient_filter", "key")

    patient_id: str | None = None
    key: str | None = None

    @property
    def effect_payload(self) -> dict[str, Any]:
        """The payload of the effect."""
        return {"patient": self.patient_id, "patient_filter": self.patient_filter, "key": self.key}



import json

from canvas_sdk.effects import Effect, EffectType
from canvas_sdk.events import EventType
from canvas_sdk.protocols import BaseProtocol
from logger import log


# Inherit from BaseProtocol to properly get registered for events
class Protocol(BaseProtocol):
    """You should put a helpful description of this protocol's behavior here."""

    # Name the event type you wish to run in response to
    RESPONDS_TO = EventType.Name(EventType.ASSESS_COMMAND__CONDITION_SELECTED)

    NARRATIVE_STRING = "I was inserted from my plugin's protocol."

    def compute(self) -> list[Effect]:
        """This method gets called when an event of the type RESPONDS_TO is fired."""
        # This class is initialized with several pieces of information you can
        # access.
        #
        # `self.event` is the event object that caused this method to be
        # called.
        #
        # `self.event.target.id` is an identifier for the object that is the subject of
        # the event. In this case, it would be the identifier of the assess
        # command. If this was a patient create event, it would be the
        # identifier of the patient. If this was a task update event, it would
        # be the identifier of the task. Etc, etc.
        # If the targeted model is already supported by the SDK,
        # you can retrieve the instance using `self.event.target.instance`
        #
        # `self.event.context` is a python dictionary of additional data that was
        # given with the event. The information given here depends on the
        # event type.
        #
        # `self.secrets` is a python dictionary of the secrets you defined in
        # your CANVAS_MANIFEST.json and set values for in the uploaded
        # plugin's configuration page: <emr_base_url>/admin/plugin_io/plugin/<plugin_id>/change/
        # Example: self.secrets['WEBHOOK_URL']

        # You can log things and see them using the Canvas CLI's log streaming
        # function.
        log.info(self.NARRATIVE_STRING)

        # Craft a payload to be returned with the effect(s).
        payload = {
            "note": {"uuid": self.context["note"]["uuid"]},
            "data": {"narrative": self.NARRATIVE_STRING},
        }

        # Return zero, one, or many effects.
        # Example:
        return [Effect(type=EffectType.LOG, payload=json.dumps(payload))]



from canvas_generated.messages.events_pb2 import Event as EventRequest
from canvas_generated.messages.events_pb2 import EventResponse, EventType

from .base import Event

__all__ = ("EventRequest", "EventResponse", "EventType", "Event")



from django.db import models

from canvas_sdk.v1.data.base import CommittableModelManager, ValueSetLookupQuerySet
from canvas_sdk.v1.data.patient import Patient
from canvas_sdk.v1.data.user import CanvasUser


class ObservationQuerySet(ValueSetLookupQuerySet):
    """ObservationQuerySet."""

    pass


class Observation(models.Model):
    """Observation."""

    class Meta:
        managed = False
        app_label = "canvas_sdk"
        db_table = "canvas_sdk_data_api_observation_001"

    objects = CommittableModelManager.from_queryset(ObservationQuerySet)()

    id = models.UUIDField()
    dbid = models.BigIntegerField(primary_key=True)
    created = models.DateTimeField()
    modified = models.DateTimeField()
    originator = models.ForeignKey(CanvasUser, on_delete=models.DO_NOTHING)
    committer = models.ForeignKey(CanvasUser, on_delete=models.DO_NOTHING)
    entered_in_error = models.ForeignKey(CanvasUser, on_delete=models.DO_NOTHING)
    deleted = models.BooleanField()
    patient = models.ForeignKey(Patient, on_delete=models.DO_NOTHING, related_name="observations")
    is_member_of = models.ForeignKey(
        "self", on_delete=models.DO_NOTHING, null=True, related_name="members"
    )
    category = models.CharField()
    units = models.TextField()
    value = models.TextField()
    note_id = models.BigIntegerField()
    name = models.TextField()
    effective_datetime = models.DateTimeField()


class ObservationCoding(models.Model):
    """ObservationCoding."""

    class Meta:
        managed = False
        app_label = "canvas_sdk"
        db_table = "canvas_sdk_data_api_observationcoding_001"

    dbid = models.BigIntegerField(primary_key=True)
    system = models.CharField()
    version = models.CharField()
    code = models.CharField()
    display = models.CharField()
    user_selected = models.BooleanField()
    observation = models.ForeignKey(
        Observation, on_delete=models.DO_NOTHING, related_name="codings"
    )


class ObservationComponent(models.Model):
    """ObservationComponent."""

    class Meta:
        managed = False
        app_label = "canvas_sdk"
        db_table = "canvas_sdk_data_api_observationcomponent_001"

    dbid = models.BigIntegerField(primary_key=True)
    created = models.DateTimeField()
    modified = models.DateTimeField()
    observation = models.ForeignKey(
        Observation, on_delete=models.DO_NOTHING, related_name="components"
    )
    value_quantity = models.TextField()
    value_quantity_unit = models.TextField()
    name = models.TextField()


class ObservationComponentCoding(models.Model):
    """ObservationComponentCoding."""

    class Meta:
        managed = False
        app_label = "canvas_sdk"
        db_table = "canvas_sdk_data_api_observationcomponentcoding_001"

    dbid = models.BigIntegerField(primary_key=True)
    system = models.CharField()
    version = models.CharField()
    code = models.CharField()
    display = models.CharField()
    user_selected = models.BooleanField()
    observation_component = models.ForeignKey(
        ObservationComponent, on_delete=models.DO_NOTHING, related_name="codings"
    )


class ObservationValueCoding(models.Model):
    """ObservationValueCoding."""

    class Meta:
        managed = False
        app_label = "canvas_sdk"
        db_table = "canvas_sdk_data_api_observationvaluecoding_001"

    dbid = models.BigIntegerField(primary_key=True)
    system = models.CharField()
    version = models.CharField()
    code = models.CharField()
    display = models.CharField()
    user_selected = models.BooleanField()
    observation = models.ForeignKey(
        Observation, on_delete=models.DO_NOTHING, related_name="value_codings"
    )



# ruff: noqa
# register all models in the app so they can be used with apps.get_model()
from canvas_sdk.v1.data.command import Command
from canvas_sdk.v1.data.condition import Condition
from canvas_sdk.v1.data.note import Note
from canvas_sdk.v1.data.patient import Patient
from canvas_sdk.v1.data.user import CanvasUser



from pathlib import Path

import pytest

import canvas_cli.main
from canvas_cli.utils.context import context


@pytest.fixture(autouse=True)
def monkeypatch_app_dir(tmp_path: Path, monkeypatch: pytest.MonkeyPatch) -> None:
    """Monkeypatch `get_app_dir` in order to return a temp dir when testing, so we don't overwrite our config file."""

    def app_dir() -> str:
        return str(tmp_path)

    monkeypatch.setattr(canvas_cli.main, "get_app_dir", app_dir)


@pytest.fixture(autouse=True)
def reset_context_variables() -> None:
    """Reset the context properties to their default value.

    This is needed because we cannot build a `reset` method in the CLIContext class,
    because `load_from_file` loads properties dynamically.
    Also since this is a CLI, it's not expected to keep the global context in memory for more than a run,
    which definitely happens with tests run
    """
    context._default_host = None



from django.db import models

from canvas_sdk.v1.data.patient import Patient
from canvas_sdk.v1.data.user import CanvasUser


class Command(models.Model):
    """Command."""

    class Meta:
        managed = False
        app_label = "canvas_sdk"
        db_table = "canvas_sdk_data_commands_command_001"

    id = models.UUIDField()
    dbid = models.BigIntegerField(primary_key=True)
    created = models.DateTimeField()
    modified = models.DateTimeField()
    originator = models.ForeignKey(CanvasUser, on_delete=models.DO_NOTHING)
    committer = models.ForeignKey(CanvasUser, on_delete=models.DO_NOTHING)
    entered_in_error = models.ForeignKey(CanvasUser, on_delete=models.DO_NOTHING)
    state = models.CharField()
    patient = models.ForeignKey(Patient, on_delete=models.DO_NOTHING)
    note_id = models.BigIntegerField()
    schema_key = models.TextField()
    data = models.JSONField()
    origination_source = models.CharField()



from enum import StrEnum
from typing import Any

from canvas_sdk.effects import EffectType, _BaseEffect


class LaunchModalEffect(_BaseEffect):
    """An Effect that will launch a modal."""

    class Meta:
        effect_type = EffectType.LAUNCH_MODAL

    class TargetType(StrEnum):
        DEFAULT_MODAL = "default_modal"
        NEW_WINDOW = "new_window"
        RIGHT_CHART_PANE = "right_chart_pane"

    url: str
    target: TargetType = TargetType.DEFAULT_MODAL

    @property
    def values(self) -> dict[str, Any]:
        """The LaunchModalEffect values."""
        return {"url": self.url, "target": self.target.value}



from .billing import BillingLineItem
from .condition import Condition, ConditionCoding
from .medication import Medication, MedicationCoding
from .organization import Organization
from .patient import Patient
from .practicelocation import PracticeLocation
from .staff import Staff
from .task import Task, TaskComment, TaskLabel

__all__ = (
    "BillingLineItem",
    "Condition",
    "ConditionCoding",
    "Medication",
    "MedicationCoding",
    "Organization",
    "Patient",
    "PracticeLocation",
    "Staff",
    "Task",
    "TaskComment",
    "TaskLabel",
)



import configparser
from datetime import datetime, timedelta
from pathlib import Path
from urllib.parse import urlparse

import keyring
import requests

from canvas_sdk.utils import Http

# Keyring namespace we'll use
KEYRING_SERVICE = __name__

CONFIG_PATH = Path.home() / ".canvas" / "credentials.ini"

LOCALHOST = "http://localhost:8000"


def get_password(username: str) -> str | None:
    """Return the stored password for username, or None."""
    return keyring.get_password(KEYRING_SERVICE, username)


def set_password(username: str, password: str) -> None:
    """Set the password for the given username."""
    keyring.set_password(KEYRING_SERVICE, username=username, password=password)


def delete_password(username: str) -> None:
    """Delete the password for the given username."""
    keyring.delete_password(KEYRING_SERVICE, username=username)


def get_config() -> configparser.ConfigParser:
    """Reads the config file and returns a ConfigParser object."""
    config = configparser.ConfigParser()
    if not config.read(CONFIG_PATH):
        raise Exception(
            f"""Please add your configuration file at '{CONFIG_PATH}' with the following format:

            [my-canvas-subdomain]
            client_id=myclientid
            client_secret=myclientsecret

            [my-dev-canvas-subdomain]
            client_id=devclientid
            client_secret=devclientsecret
            is_default=true

            [localhost]
            client_id=localclientid
            client_secret=localclientsecret
            """
        )
    return config


def read_config(host: str, property: str) -> str:
    """Reads the config file and returns the property for a given section."""
    config = get_config()
    if host not in config:
        raise Exception(f"'{host}' is not found in the configuration file at '{CONFIG_PATH}'")
    return config.get(host, property)


def get_api_client_credentials(host: str) -> str:
    """Either return the given api_key, or fetch it from the keyring."""
    hostname = urlparse(host).hostname

    if not hostname:
        raise ValueError("Could not parse hostname from URL")

    instance = hostname.removesuffix(".canvasmedical.com")

    client_id = read_config(instance, "client_id")
    client_secret = read_config(instance, "client_secret")

    return f"client_id={client_id}&client_secret={client_secret}"


def get_default_host(host: str | None = None) -> str:
    """Return the explicitly stated default host, or first if none is indicated."""
    if host:
        if "://" in host:
            return host

        if "localhost" in host:
            return LOCALHOST

        return f"https://{host}.canvasmedical.com"

    config = get_config()
    if not (hosts := config.sections()):
        raise Exception(f"No hosts found in the configuration file at '{CONFIG_PATH}'")

    first_default_host = next(
        (host for host in hosts if config.getboolean(host, "is_default", fallback=False) is True),
        hosts[0],
    )
    if first_default_host == "localhost":
        return LOCALHOST

    return f"https://{first_default_host}.canvasmedical.com"


def request_api_token(host: str, api_client_credentials: str) -> dict:
    """Request an api token using the provided client_id and client_secret."""
    grant_type = "grant_type=client_credentials"
    scope = "scope=system/Plugins.*"

    http = Http()
    token_response = http.post(
        f"{host}/auth/token/",
        headers={"Content-Type": "application/x-www-form-urlencoded"},
        data=f"{grant_type}&{scope}&{api_client_credentials}",
    )
    if token_response.status_code != requests.codes.ok:
        raise Exception(f"Unable to get a valid access token from the given host '{host}'")
    return token_response.json()


def is_token_valid(host_token_key: str, expiration_date: datetime | None = None) -> bool:
    """True if the token has not expired yet."""
    token_exp_date_key = f"{host_token_key}|exp_date"

    if expiration_date:
        if expiration_date <= datetime.now():
            return False
        set_password(token_exp_date_key, expiration_date.isoformat())
        return True

    stored_expiration_date = get_password(token_exp_date_key)
    return (
        stored_expiration_date is not None
        and datetime.fromisoformat(stored_expiration_date) > datetime.now()
    )


def get_or_request_api_token(host: str | None = None) -> str:
    """Returns an existing stored token if it has not expired, or requests a new one."""
    if not (host := get_default_host(host)):
        raise Exception(
            f"Please specify a host or add one to the configuration file at '{CONFIG_PATH}'"
        )

    host_token_key = f"{host}|token"
    token = get_password(host_token_key)

    if token and is_token_valid(host_token_key):
        return token

    api_client_credentials = get_api_client_credentials(host)

    if not (token_response := request_api_token(host, api_client_credentials)):
        raise Exception(f"A token could not be acquired from the given host '{host}'")

    token_expiration_date = datetime.now() + timedelta(seconds=token_response["expires_in"])
    if not is_token_valid(host_token_key, token_expiration_date):
        raise Exception(f"A valid token could not be acquired from the given host '{host}'")

    new_token = token_response["access_token"]
    set_password(host_token_key, new_token)
    return new_token



import ast
import builtins
from _ast import AnnAssign
from functools import cached_property
from typing import Any, cast

from RestrictedPython import (
    CompileResult,
    PrintCollector,
    RestrictingNodeTransformer,
    compile_restricted_exec,
    safe_builtins,
    utility_builtins,
)
from RestrictedPython.Eval import default_guarded_getitem
from RestrictedPython.Guards import (
    guarded_iter_unpack_sequence,
    guarded_unpack_sequence,
)
from RestrictedPython.transformer import (
    ALLOWED_FUNC_NAMES,
    FORBIDDEN_FUNC_NAMES,
    copy_locations,
)

##
# ALLOWED_MODULES
#
# The modules in this list are the only ones that can be imported in a sandboxed
# runtime.
#
ALLOWED_MODULES = frozenset(
    [
        "_strptime",
        "arrow",
        "base64",
        "cached_property",
        "canvas_sdk.commands",
        "canvas_sdk.data",
        "canvas_sdk.effects",
        "canvas_sdk.events",
        "canvas_sdk.handlers",
        "canvas_sdk.protocols",
        "canvas_sdk.utils",
        "canvas_sdk.v1",
        "canvas_sdk.value_set",
        "canvas_sdk.views",
        "contextlib",
        "datetime",
        "dateutil",
        "django.db.models",
        "django.utils.functional",
        "enum",
        "functools",
        "hashlib",
        "hmac",
        "http",
        "json",
        "logger",
        "math",
        "operator",
        "pickletools",
        "random",
        "rapidfuzz",
        "re",
        "requests",
        "string",
        "time",
        "traceback",
        "typing",
        "urllib",
        "uuid",
    ]
)


def _is_known_module(name: str) -> bool:
    return any(name.startswith(m) for m in ALLOWED_MODULES)


def _unrestricted(_ob: Any, *args: Any, **kwargs: Any) -> Any:
    """Return the given object, unmodified."""
    return _ob


def _apply(_ob: Any, *args: Any, **kwargs: Any) -> Any:
    """Call the bound method with args, support calling super().__init__()."""
    return _ob(*args, **kwargs)


class Sandbox:
    """A restricted sandbox for safely executing arbitrary Python code."""

    source_code: str
    namespace: str

    class Transformer(RestrictingNodeTransformer):
        """A node transformer for customizing the sandbox compiler."""

        def visit_AnnAssign(self, node: AnnAssign) -> AnnAssign:
            """Allow type annotations."""
            return node

        def check_import_names(self, node: ast.ImportFrom) -> ast.AST:
            """Check the names being imported.

            This is a protection against rebinding dunder names like
            _getitem_, _write_ via imports.

            => 'from _a import x' is ok, because '_a' is not added to the scope.
            """
            for name in node.names:
                if "*" in name.name and node.module and not _is_known_module(node.module):
                    self.error(node, '"*" imports are not allowed.')
                self.check_name(node, name.name)
                if name.asname:
                    self.check_name(node, name.asname)

            return self.node_contents_visit(node)

        def check_name(
            self, node: ast.ImportFrom, name: str | None, allow_magic_methods: bool = False
        ) -> None:
            """Check names if they are allowed.

            If ``allow_magic_methods is True`` names in `ALLOWED_FUNC_NAMES`
            are additionally allowed although their names start with `_`.

            Override to turn errors into warnings for leading underscores.
            """
            if name is None:
                return

            if (
                name.startswith("_")
                and name != "_"
                and not (
                    allow_magic_methods and name in ALLOWED_FUNC_NAMES and node.col_offset != 0
                )
            ):
                self.warn(
                    node,
                    f'"{name}" is an invalid variable name because it ' 'starts with "_"',
                )
            elif name.endswith("__roles__"):
                self.error(
                    node,
                    f'"{name}" is an invalid variable name because ' 'it ends with "__roles__".',
                )
            elif name in FORBIDDEN_FUNC_NAMES:
                self.error(node, f'"{name}" is a reserved name.')

        def visit_Attribute(self, node: ast.Attribute) -> ast.AST:
            """Checks and mutates attribute access/assignment.

            'a.b' becomes '_getattr_(a, "b")'
            'a.b = c' becomes '_write_(a).b = c'
            'del a.b' becomes 'del _write_(a).b'

            The _write_ function should return a security proxy.

            Override to turn errors into warnings for leading underscores.
            """
            if node.attr.startswith("_") and node.attr != "_":
                self.warn(
                    node,
                    f'"{node.attr}" is an invalid attribute name because it starts ' 'with "_".',
                )

            if node.attr.endswith("__roles__"):
                self.error(
                    node,
                    f'"{node.attr}" is an invalid attribute name because it ends '
                    'with "__roles__".',
                )

            if isinstance(node.ctx, ast.Load):
                node = self.node_contents_visit(node)
                new_node = ast.Call(
                    func=ast.Name("_getattr_", ast.Load()),
                    args=[node.value, ast.Str(node.attr)],
                    keywords=[],
                )

                copy_locations(new_node, node)
                return new_node

            elif isinstance(node.ctx, ast.Store | ast.Del):
                node = self.node_contents_visit(node)
                new_value = ast.Call(
                    func=ast.Name("_write_", ast.Load()), args=[node.value], keywords=[]
                )

                copy_locations(new_value, node.value)
                node.value = new_value
                return node

            else:  # pragma: no cover
                # Impossible Case only ctx Load, Store and Del are defined in ast.
                raise NotImplementedError(f"Unknown ctx type: {type(node.ctx)}")

    def __init__(self, source_code: str, namespace: str | None = None) -> None:
        if source_code is None:
            raise TypeError("source_code may not be None")
        self.namespace = namespace or "protocols"
        self.source_code = source_code

    @cached_property
    def package_name(self) -> str | None:
        """Return the root package name."""
        return self.namespace.split(".")[0] if self.namespace else None

    @cached_property
    def scope(self) -> dict[str, Any]:
        """Return the scope used for evaluation."""
        return {
            "__builtins__": {
                **safe_builtins.copy(),
                **utility_builtins.copy(),
                "__import__": self._safe_import,
                "classmethod": builtins.classmethod,
                "staticmethod": builtins.staticmethod,
                "any": builtins.any,
                "all": builtins.all,
                "enumerate": builtins.enumerate,
                "property": builtins.property,
                "super": builtins.super,
                "dict": builtins.dict,
                "filter": builtins.filter,
                "max": builtins.max,
                "min": builtins.min,
                "list": builtins.list,
                "next": builtins.next,
                "iter": builtins.iter,
                "type": builtins.type,
            },
            "__metaclass__": type,
            "__name__": self.namespace,
            "_write_": _unrestricted,
            "_getiter_": _unrestricted,
            "_getitem_": default_guarded_getitem,
            "_getattr_": getattr,
            "_print_": PrintCollector,
            "_apply_": _apply,
            "_inplacevar_": _unrestricted,
            "_iter_unpack_sequence_": guarded_iter_unpack_sequence,
            "_unpack_sequence_": guarded_unpack_sequence,
            "hasattr": hasattr,
        }

    @cached_property
    def compile_result(self) -> CompileResult:
        """Compile the source code into bytecode."""
        return compile_restricted_exec(self.source_code, policy=self.Transformer)

    @property
    def errors(self) -> tuple[str, ...]:
        """Return errors encountered when compiling the source code."""
        return cast(tuple[str, ...], self.compile_result.errors)

    @property
    def warnings(self) -> tuple[str, ...]:
        """Return warnings encountered when compiling the source code."""
        return cast(tuple[str, ...], self.compile_result.warnings)

    def _is_known_module(self, name: str) -> bool:
        return bool(
            _is_known_module(name)
            or (self.package_name and name.split(".")[0] == self.package_name)
        )

    def _safe_import(self, name: str, *args: Any, **kwargs: Any) -> Any:
        if not (self._is_known_module(name)):
            raise ImportError(f"{name!r} is not an allowed import.")
        return __import__(name, *args, **kwargs)

    def execute(self) -> dict:
        """Execute the given code in a restricted sandbox."""
        if self.errors:
            raise RuntimeError(f"Code is invalid: {self.errors}")

        exec(self.compile_result.code, self.scope)

        return self.scope



from typing import Any

import pytest
from pydantic import ValidationError

from canvas_sdk.commands import (
    AssessCommand,
    DiagnoseCommand,
    GoalCommand,
    HistoryOfPresentIllnessCommand,
    MedicationStatementCommand,
    PlanCommand,
    PrescribeCommand,
    QuestionnaireCommand,
    ReasonForVisitCommand,
    StopMedicationCommand,
    UpdateGoalCommand,
)
from canvas_sdk.commands.base import _BaseCommand
from canvas_sdk.effects.protocol_card import ProtocolCard, Recommendation


def test_apply_method_succeeds_with_patient_id_and_key() -> None:
    """Test that the apply method succeeds with all required fields."""
    p = ProtocolCard(patient_id="uuid", key="something-unique")
    applied = p.apply()
    assert (
        applied.payload
        == '{"patient": "uuid", "patient_filter": null, "key": "something-unique", "data": {"title": "", "narrative": "", "recommendations": [], "status": "due", "feedback_enabled": false, "due_in": -1}}'
    )


def test_apply_method_raises_error_without_patient_id_and_key() -> None:
    """Test that the apply method raises an error when missing required fields."""
    p = ProtocolCard()

    with pytest.raises(ValidationError) as e:
        p.apply()
    err_msg = repr(e.value)

    assert "2 validation errors for ProtocolCard" in err_msg
    assert (
        "Field 'patient_id' or 'patient_filter' is required to apply a ProtocolCard [type=missing, input_value=None, input_type=NoneType]"
        in err_msg
    )
    assert (
        "Field 'key' is required to apply a ProtocolCard [type=missing, input_value=None, input_type=NoneType]"
        in err_msg
    )


@pytest.mark.parametrize(
    "init_params,rec1_params,rec2_params",
    [
        (
            {
                "key": "link_rec",
                "patient_id": "patientuuid",
                "title": "This is a test!",
                "narrative": "we should only expect a link and a button",
            },
            {
                "title": "this is a link",
                "button": "click this",
                "href": "https://canvasmedical.com/",
            },
            {"title": "second link", "button": "don't click this", "href": "https://google.com/"},
        ),
        (
            {
                "key": "command_rec",
                "patient_id": "patientuuid",
                "title": "This is a test for command recommendations!",
                "narrative": "we should only expect buttons to insert commands",
            },
            {
                "title": "this is a command",
                "button": "click this",
                "command": "updategoal",
                "context": {"progress": "none"},
            },
            {
                "title": "another command",
                "button": "hypertension",
                "command": "diagnose",
                "context": {"background": "stuff"},
            },
        ),
        (
            {
                "patient_id": "patientuuid",
                "key": "command_rec_with_coding_filter",
                "title": "This is a test for command recommendations with coding filters!",
                "narrative": "we should only expect buttons to insert commands",
            },
            {
                "title": "hypertension",
                "button": "diagnose",
                "command": "diagnose",
                "context": {"background": "hey", "icd10_code": "I10"},
            },
            {
                "title": "fake medication",
                "button": "prescribe",
                "command": "prescribe",
                "context": {"sig": "1pobid"},
            },
        ),
    ],
)
def test_add_recommendations(
    init_params: dict[Any, Any], rec1_params: dict[Any, Any], rec2_params: dict[Any, Any]
) -> None:
    """Test that the add_recommendation method adds recommendations to the ProtocolCard."""
    p = ProtocolCard(**init_params)
    p.add_recommendation(**rec1_params)
    p.recommendations.append(Recommendation(**rec2_params))
    assert p.values == {
        "title": init_params["title"],
        "narrative": init_params["narrative"],
        "recommendations": [
            {
                "title": rec1_params.get("title"),
                "button": rec1_params.get("button"),
                "href": rec1_params.get("href"),
                "command": {"type": rec1_params["command"]} if "command" in rec1_params else {},
                "context": rec1_params.get("context", {}),
                "key": 0,
            },
            {
                "title": rec2_params.get("title"),
                "button": rec2_params.get("button"),
                "href": rec2_params.get("href"),
                "command": {"type": rec2_params["command"]} if "command" in rec2_params else {},
                "context": rec2_params.get("context", {}),
                "key": 1,
            },
        ],
        "status": "due",
        "due_in": -1,
        "feedback_enabled": False,
    }


@pytest.mark.parametrize(
    "Command,init_params",
    [
        (AssessCommand, {}),
        (DiagnoseCommand, {"icd10_code": "I10"}),
        (GoalCommand, {}),
        (HistoryOfPresentIllnessCommand, {}),
        (MedicationStatementCommand, {"fdb_code": "fakeroo"}),
        (PlanCommand, {}),
        (PrescribeCommand, {"fdb_code": "fake"}),
        (QuestionnaireCommand, {}),
        (ReasonForVisitCommand, {}),
        (StopMedicationCommand, {}),
        (UpdateGoalCommand, {}),
    ],
)
def test_add_recommendations_from_commands(
    Command: type[_BaseCommand], init_params: dict[str, str]
) -> None:
    """Test that the add_recommendation method adds recommendations from commands to the ProtocolCard."""
    cmd = Command(**init_params)
    p = ProtocolCard(patient_id="uuid", key="commands")
    p.recommendations.append(cmd.recommend())
    p.recommendations.append(cmd.recommend(title="hello", button="click"))
    p.add_recommendation(
        title="yeehaw", button="click here", command=cmd.Meta.key.lower(), context=init_params
    )

    rec1, rec2, rec3 = p.values["recommendations"]
    assert rec1["title"] == ""
    assert rec1["button"] == cmd.constantized_key().lower().replace("_", " ")
    assert rec1["href"] is None
    assert rec1["context"] == cmd.values
    assert rec1["command"]["type"] == cmd.Meta.key.lower()

    assert rec2["title"] == "hello"
    assert rec2["button"] == "click"
    assert rec2["href"] is None
    assert rec2["context"] == cmd.values
    assert rec2["command"]["type"] == cmd.Meta.key.lower()

    assert rec3["title"] == "yeehaw"
    assert rec3["button"] == "click here"
    assert rec3["href"] is None
    assert rec3["command"]["type"] == cmd.Meta.key.lower()
    for k, v in init_params.items():
        assert rec3["context"][k] == v



import shutil
from collections.abc import Generator
from contextlib import chdir
from datetime import datetime
from pathlib import Path
from typing import Any

import pytest
import requests
from django.core.exceptions import ImproperlyConfigured
from pydantic import ValidationError
from typer.testing import CliRunner

import settings
from canvas_cli.apps.plugin.plugin import _build_package, plugin_url
from canvas_cli.main import app
from canvas_sdk.commands.tests.test_utils import MaskedValue
from canvas_sdk.effects.banner_alert import AddBannerAlert, RemoveBannerAlert

runner = CliRunner()


@pytest.fixture(scope="session")
def token() -> MaskedValue:
    """Get a valid token."""
    return MaskedValue(
        requests.post(
            f"{settings.INTEGRATION_TEST_URL}/auth/token/",
            headers={"Content-Type": "application/x-www-form-urlencoded"},
            data={
                "grant_type": "client_credentials",
                "client_id": settings.INTEGRATION_TEST_CLIENT_ID,
                "client_secret": settings.INTEGRATION_TEST_CLIENT_SECRET,
            },
        ).json()["access_token"]
    )


@pytest.fixture(scope="session")
def first_patient_id(token: MaskedValue) -> dict:
    """Get the first patient id."""
    headers = {
        "Authorization": f"Bearer {token.value}",
        "Content-Type": "application/json",
        "Accept": "application/json",
    }
    patients = requests.get(f"{settings.INTEGRATION_TEST_URL}/api/Patient", headers=headers).json()
    return patients["entry"][0]["resource"]["key"]


@pytest.fixture(scope="session")
def plugin_name() -> str:
    """Get the plugin name to be used."""
    return f"addbanneralert{datetime.now().timestamp()}".replace(".", "")


@pytest.fixture(scope="session")
def write_and_install_protocol_and_clean_up(
    first_patient_id: str, plugin_name: str, token: MaskedValue
) -> Generator[Any, Any, Any]:
    """Write the protocol code, install the plugin, and clean up after the test."""
    if not settings.INTEGRATION_TEST_URL:
        raise ImproperlyConfigured("INTEGRATION_TEST_URL is not set")

    # write the protocol
    with chdir(Path("./custom-plugins")):
        runner.invoke(app, "init", input=plugin_name)

    protocol_code = f"""
from canvas_sdk.effects.banner_alert import AddBannerAlert
from canvas_sdk.events import EventType
from canvas_sdk.protocols import BaseProtocol

class Protocol(BaseProtocol):
    RESPONDS_TO = EventType.Name(EventType.ENCOUNTER_CREATED)
    def compute(self):
        return [
            AddBannerAlert(
                patient_id="{first_patient_id}",
                key="{plugin_name}",
                narrative="this is a test",
                placement=[AddBannerAlert.Placement.CHART],
                intent=AddBannerAlert.Intent.INFO,
            ).apply()
        ]
"""

    with open(f"./custom-plugins/{plugin_name}/protocols/my_protocol.py", "w") as protocol:
        protocol.write(protocol_code)

    with open(_build_package(Path(f"./custom-plugins/{plugin_name}")), "rb") as package:
        # install the plugin
        requests.post(
            plugin_url(settings.INTEGRATION_TEST_URL),
            data={"is_enabled": True},
            files={"package": package},
            headers={"Authorization": f"Bearer {token.value}"},
        )

    yield

    # clean up
    if Path(f"./custom-plugins/{plugin_name}").exists():
        shutil.rmtree(Path(f"./custom-plugins/{plugin_name}"))

    # disable
    requests.patch(
        plugin_url(settings.INTEGRATION_TEST_URL, plugin_name),
        data={"is_enabled": False},
        headers={
            "Authorization": f"Bearer {token.value}",
        },
    )
    # delete
    requests.delete(
        plugin_url(settings.INTEGRATION_TEST_URL, plugin_name),
        headers={"Authorization": f"Bearer {token.value}"},
    )

    # confirm no more banner
    patient_banners_none = requests.get(
        f"{settings.INTEGRATION_TEST_URL}/api/BannerAlert/?patient__key={first_patient_id}",
        headers={
            "Authorization": f"Bearer {token.value}",
        },
    ).json()
    patient_banner = next(
        (b for b in patient_banners_none["results"] if b["key"] == plugin_name), None
    )
    assert patient_banner is None


@pytest.mark.integtest
def test_protocol_that_adds_banner_alert(
    write_and_install_protocol_and_clean_up: None,
    token: MaskedValue,
    plugin_name: str,
    first_patient_id: str,
) -> None:
    """Test that the protocol adds a banner alert."""
    # trigger the event
    requests.post(
        f"{settings.INTEGRATION_TEST_URL}/api/Note/",
        headers={
            "Authorization": f"Bearer {token.value}",
            "Content-Type": "application/json",
            "Accept": "application/json",
        },
        json={
            "patient": 1,
            "provider": 1,
            "note_type": "office",
            "note_type_version": 1,
            "lastModifiedBySessionKey": "8fee3c03a525cebee1d8a6b8e63dd4dg",
        },
    )

    patient_banners = requests.get(
        f"{settings.INTEGRATION_TEST_URL}/api/BannerAlert/?patient__key={first_patient_id}",
        headers={
            "Authorization": f"Bearer {token.value}",
        },
    ).json()
    assert patient_banners["count"] > 0

    patient_banner = next(b for b in patient_banners["results"] if b["key"] == plugin_name)
    assert patient_banner["pluginName"] == plugin_name
    assert patient_banner["narrative"] == "this is a test"
    assert patient_banner["placement"] == ["chart"]
    assert patient_banner["intent"] == "info"
    assert patient_banner["href"] is None
    assert patient_banner["status"] == "active"


@pytest.mark.parametrize(
    "Effect,params,expected_payload",
    [
        (
            AddBannerAlert,
            {
                "patient_id": "uuid",
                "key": "test-key",
                "narrative": "hellooo",
                "placement": [AddBannerAlert.Placement.APPOINTMENT_CARD],
                "intent": AddBannerAlert.Intent.INFO,
            },
            '{"patient": "uuid", "patient_filter": null, "key": "test-key", "data": {"narrative": "hellooo", "placement": ["appointment_card"], "intent": "info", "href": null}}',
        ),
        (
            AddBannerAlert,
            {
                "patient_filter": {"active": True},
                "key": "test-key",
                "narrative": "hellooo",
                "placement": [AddBannerAlert.Placement.APPOINTMENT_CARD],
                "intent": AddBannerAlert.Intent.INFO,
            },
            '{"patient": null, "patient_filter": {"active": true}, "key": "test-key", "data": {"narrative": "hellooo", "placement": ["appointment_card"], "intent": "info", "href": null}}',
        ),
        (
            RemoveBannerAlert,
            {"patient_id": "uuid", "key": "testeroo"},
            '{"patient": "uuid", "patient_filter": null, "key": "testeroo"}',
        ),
        (
            RemoveBannerAlert,
            {"patient_filter": {"active": True}, "key": "testeroo"},
            '{"patient": null, "patient_filter": {"active": true}, "key": "testeroo"}',
        ),
    ],
)
def test_banner_alert_apply_method_succeeds_with_all_required_fields(
    Effect: type[AddBannerAlert] | type[RemoveBannerAlert], params: dict, expected_payload: str
) -> None:
    """Test that the apply method succeeds with all required fields."""
    b = Effect()
    for k, v in params.items():
        setattr(b, k, v)
    applied = b.apply()
    assert applied.payload == expected_payload


@pytest.mark.parametrize(
    "Effect,expected_err_msgs",
    [
        (
            AddBannerAlert,
            [
                "5 validation errors for AddBannerAlert",
                "Field 'patient_id' or 'patient_filter' is required to apply an AddBannerAlert [type=missing",
                "Field 'key' is required to apply an AddBannerAlert [type=missing",
                "Field 'narrative' is required to apply an AddBannerAlert [type=missing",
                "Field 'placement' is required to apply an AddBannerAlert [type=missing",
                "Field 'intent' is required to apply an AddBannerAlert [type=missing",
            ],
        ),
        (
            RemoveBannerAlert,
            [
                "2 validation errors for RemoveBannerAlert",
                "Field 'patient_id' or 'patient_filter' is required to apply a RemoveBannerAlert [type=missing",
                "Field 'key' is required to apply a RemoveBannerAlert [type=missing",
            ],
        ),
    ],
)
def test_banner_alert_apply_method_raises_error_without_required_fields(
    Effect: type[AddBannerAlert] | type[RemoveBannerAlert], expected_err_msgs: str
) -> None:
    """Test that the apply method raises an error when missing required fields."""
    b = Effect()
    with pytest.raises(ValidationError) as e:
        b.apply()
    err_msg = repr(e.value)
    for expected in expected_err_msgs:
        assert expected in err_msg



from typing import Literal

from pydantic_core import InitErrorDetails

from canvas_sdk.commands.base import _BaseCommand
from canvas_sdk.commands.constants import Coding


class ReasonForVisitCommand(_BaseCommand):
    """A class for managing a ReasonForVisit command within a specific note."""

    class Meta:
        key = "reasonForVisit"

    structured: bool = False
    # how do we make sure that coding is a valid rfv coding from their home-app?
    coding: Coding | None = None
    comment: str | None = None

    def _get_error_details(
        self, method: Literal["originate", "edit", "delete", "commit", "enter_in_error"]
    ) -> list[InitErrorDetails]:
        errors = super()._get_error_details(method)
        if self.structured and not self.coding:
            errors.append(
                self._create_error_detail(
                    "value", "Structured RFV should have a coding.", self.coding
                )
            )
        return errors

    @classmethod
    def command_schema(cls) -> dict:
        """The schema of the command."""
        command_schema = super().command_schema()
        # the commands api does not include the 'structured' field in the fields response
        command_schema.pop("structured")
        return command_schema

    @property
    def values(self) -> dict:
        """The ReasonForVisit command's field values."""
        return {"structured": self.structured, "coding": self.coding, "comment": self.comment}



import json
from typing import cast
from urllib.parse import urlparse

import typer
import websocket

from canvas_cli.apps.auth.utils import get_default_host, get_or_request_api_token


def _on_message(ws: websocket.WebSocket, message: str) -> None:
    message_to_print = message
    try:
        message_json = json.loads(message)
        message_to_print = f"{message_json['timestamp']} | {message_json['message']}"
    except ValueError:
        pass
    print(message_to_print)


def _on_error(ws: websocket.WebSocket, error: str) -> None:
    print(f"Error: {error}")


def _on_close(ws: websocket.WebSocket, close_status_code: str, close_msg: str) -> None:
    print(f"Connection closed with status code {close_status_code}: {close_msg}")


def _on_open(ws: websocket.WebSocket) -> None:
    print("Connected to the logging service")


def logs(
    host: str | None = typer.Option(
        callback=get_default_host, help="Canvas instance to connect to", default=None
    ),
) -> None:
    """Listens and prints log streams from the instance."""
    if not host:
        raise typer.BadParameter("Please specify a host or add one to the configuration file")

    token = get_or_request_api_token(host)

    # Resolve the instance name from the Canvas host URL (e.g., extract
    # 'example' from 'https://example.canvasmedical.com/')
    hostname = cast(str, urlparse(host).hostname)
    instance = hostname.removesuffix(".canvasmedical.com")

    print(
        "Connecting to the log stream. Please be patient as there may be a delay before log messages appear."
    )
    websocket_uri = f"wss://logs.console.canvasmedical.com/{instance}?token={token}"

    try:
        ws = websocket.WebSocketApp(
            websocket_uri,
            on_message=_on_message,
            on_error=_on_error,
            on_close=_on_close,
        )
        ws.on_open = _on_open
        ws.run_forever()

    except KeyboardInterrupt:
        raise typer.Exit(0) from None



import arrow


class Timeframe:
    """A class representing a timeframe with a start and and end."""

    def __init__(self, start: arrow.Arrow, end: arrow.Arrow):
        self.start = start
        self.end = end

    def __str__(self) -> str:
        return f"<Timeframe start={self.start}, end={self.end}>"

    @property
    def duration(self) -> int:
        """Returns the number of days in the timeframe."""
        return (self.end - self.start).days

    def increased_by(self, years: int = 0, months: int = 0, days: int = 0) -> "Timeframe":
        """Returns a new Timeframe object increased by the years, months, days in the arguments."""
        start = self.start
        end = self.end

        if years > 0:
            end = end.shift(years=years)
        elif years < 0:
            start = start.shift(years=years)

        if months > 0:
            end = end.shift(months=months)
        elif months < 0:
            start = start.shift(months=months)

        if days > 0:
            end = end.shift(days=days)
        elif days < 0:
            start = start.shift(days=days)

        return Timeframe(start=start, end=end)



import random
import shutil
import string
from contextlib import chdir
from datetime import datetime
from decimal import Decimal
from pathlib import Path
from typing import Any, cast

import pytest
import requests
from pydantic import ValidationError
from typer.testing import CliRunner

import settings
from canvas_cli.apps.plugin.plugin import _build_package, plugin_url
from canvas_cli.main import app
from canvas_sdk.commands import (
    AssessCommand,
    DiagnoseCommand,
    GoalCommand,
    HistoryOfPresentIllnessCommand,
    MedicationStatementCommand,
    PlanCommand,
    PrescribeCommand,
    QuestionnaireCommand,
    ReasonForVisitCommand,
    StopMedicationCommand,
    UpdateGoalCommand,
)
from canvas_sdk.commands.base import _BaseCommand
from canvas_sdk.commands.constants import ClinicalQuantity, Coding

runner = CliRunner()


class MaskedValue:
    """A class to mask sensitive values in tests."""

    def __init__(self, value: str) -> None:
        self.value = value

    def __repr__(self) -> str:
        return "MaskedValue(********)"

    def __str___(self) -> str:
        return "*******"


def get_field_type_unformatted(field_props: dict[str, Any]) -> str:
    """Get the unformatted field type from the field properties."""
    if t := field_props.get("type"):
        return field_props.get("format") or t

    first_in_union: dict = field_props.get("anyOf", field_props.get("allOf"))[0]
    if "$ref" in first_in_union:
        return first_in_union["$ref"].split("#/$defs/")[-1]
    return first_in_union.get("format") or first_in_union["type"]


def get_field_type(field_props: dict) -> str:
    """Get the field type from the field properties."""
    return get_field_type_unformatted(field_props).replace("-", "").replace("array", "list")


def random_string() -> str:
    """Generate a random string."""
    return "".join(random.choices(string.ascii_uppercase + string.digits, k=7))


def fake(field_props: dict, Command: type[_BaseCommand]) -> Any:
    """Generate a fake value for a field."""
    t = get_field_type(field_props)
    match t:
        case "string":
            return random_string()
        case "integer":
            return random.randint(1, 10)
        case "datetime":
            return datetime.now()
        case "boolean":
            return random.choice([True, False])
        case "number":
            return Decimal(random.randrange(1, 200))
        case "array":
            num_items = random.randint(0, 5)
            item_props = field_props["anyOf"][0]["items"]
            return [fake(item_props, Command) for i in range(num_items)]
        case "list":
            num_items = random.randint(0, field_props.get("maxItems", 5))
            item_props = field_props.get("items")
            return [fake(item_props, Command) for i in range(num_items)] if item_props else []
        case "Coding":
            return Coding(system=random_string(), code=random_string(), display=random_string())
        case "ClinicalQuantity":
            return ClinicalQuantity(representative_ndc="ndc", ncpdp_quantity_qualifier_code="code")
    if t[0].isupper():
        return random.choice(list(getattr(Command, t)))


def raises_wrong_type_error(
    Command: type[_BaseCommand],
    field: str,
) -> None:
    """Test that the correct error is raised when the wrong type is passed to a field."""
    field_props = Command.model_json_schema()["properties"][field]
    field_type = get_field_type(field_props)
    wrong_field_type = "integer" if field_type == "string" else "string"

    with pytest.raises(ValidationError) as e1:
        err_kwargs = {field: fake({"type": wrong_field_type}, Command)}
        Command(**err_kwargs)
    err_msg1 = repr(e1.value)

    valid_kwargs = {field: fake(field_props, Command)}
    cmd = Command(**valid_kwargs)
    err_value = fake({"type": wrong_field_type}, Command)
    with pytest.raises(ValidationError) as e2:
        setattr(cmd, field, err_value)
    err_msg2 = repr(e2.value)

    assert f"1 validation error for {Command.__name__}\n{field}" in err_msg1
    assert f"1 validation error for {Command.__name__}\n{field}" in err_msg2

    field_type = (
        "dictionary" if field_type == "Coding" or field_type == "ClinicalQuantity" else field_type
    )
    if field_type == "number":
        assert "Input should be an instance of Decimal" in err_msg1
        assert "Input should be an instance of Decimal" in err_msg2
    elif field_type[0].isupper():
        assert f"Input should be an instance of {Command.__name__}.{field_type}" in err_msg1
        assert f"Input should be an instance of {Command.__name__}.{field_type}" in err_msg2
    else:
        assert f"Input should be a valid {field_type}" in err_msg1
        assert f"Input should be a valid {field_type}" in err_msg2


def raises_none_error_for_effect_method(
    Command: type[_BaseCommand],
    method: str,
) -> None:
    """Test that the correct error is raised when a required field is None for an effect method."""
    cmd_name = Command.__name__
    cmd_name_article = "an" if cmd_name.startswith(("A", "E", "I", "O", "U")) else "a"

    cmd = Command()
    method_required_fields = cmd._get_effect_method_required_fields(method)
    with pytest.raises(ValidationError) as e:
        getattr(cmd, method)()
    e_msg = repr(e.value)
    missing_fields = [field for field in method_required_fields if getattr(cmd, field) is None]
    num_errs = len(missing_fields)
    assert f"{num_errs} validation error{'s' if num_errs > 1 else ''} for {cmd_name}" in e_msg

    for f in missing_fields:
        assert (
            f"Field '{f}' is required to {method.replace('_', ' ')} {cmd_name_article} {cmd_name} [type=missing, input_value=None, input_type=NoneType]"
            in e_msg
        )


def write_protocol_code(
    note_uuid: str, plugin_name: str, commands: list[type[_BaseCommand]]
) -> None:
    """Test that the protocol code is written correctly."""
    imports = ", ".join([c.__name__ for c in commands])
    effects = ", ".join([f"{c.__name__}(note_uuid='{note_uuid}').originate()" for c in commands])

    protocol_code = f"""from canvas_sdk.commands import {imports}
from canvas_sdk.events import EventType
from canvas_sdk.protocols import BaseProtocol

class Protocol(BaseProtocol):
    RESPONDS_TO = EventType.Name(EventType.ENCOUNTER_CREATED)
    def compute(self):
        return [{effects}]
"""

    with chdir(Path("./custom-plugins")):
        runner.invoke(app, "init", input=plugin_name)

    with open(f"./custom-plugins/{plugin_name}/protocols/my_protocol.py", "w") as protocol:
        protocol.write(protocol_code)


def install_plugin(plugin_name: str, token: MaskedValue) -> None:
    """Install a plugin."""
    with open(_build_package(Path(f"./custom-plugins/{plugin_name}")), "rb") as package:
        requests.post(
            plugin_url(cast(str, settings.INTEGRATION_TEST_URL)),
            data={"is_enabled": True},
            files={"package": package},
            headers={"Authorization": f"Bearer {token.value}"},
        )


def trigger_plugin_event(token: MaskedValue) -> None:
    """Trigger a plugin event."""
    requests.post(
        f"{settings.INTEGRATION_TEST_URL}/api/Note/",
        headers={
            "Authorization": f"Bearer {token.value}",
            "Content-Type": "application/json",
            "Accept": "application/json",
        },
        json={
            "patient": 2,
            "provider": 1,
            "note_type": "office",
            "note_type_version": 1,
            "lastModifiedBySessionKey": "8fee3c03a525cebee1d8a6b8e63dd4dg",
        },
    )


def get_original_note_body_commands(new_note_id: int, token: MaskedValue) -> list[str]:
    """Get the commands from the original note body."""
    original_note = requests.get(
        f"{settings.INTEGRATION_TEST_URL}/api/Note/{new_note_id}",
        headers={
            "Authorization": f"Bearer {token.value}",
            "Content-Type": "application/json",
            "Accept": "application/json",
        },
    ).json()

    body = original_note["body"]
    return [
        line["value"]
        for line in body
        if "data" in line
        and "commandUuid" in line["data"]
        and "id" in line["data"]
        and line["type"] == "command"
    ]


def clean_up_files_and_plugins(plugin_name: str, token: MaskedValue) -> None:
    """Clean up the files and plugins."""
    # clean up
    if Path(f"./custom-plugins/{plugin_name}").exists():
        shutil.rmtree(Path(f"./custom-plugins/{plugin_name}"))

    # disable
    requests.patch(
        plugin_url(cast(str, settings.INTEGRATION_TEST_URL), plugin_name),
        data={"is_enabled": False},
        headers={
            "Authorization": f"Bearer {token.value}",
        },
    )
    # delete
    requests.delete(
        plugin_url(cast(str, settings.INTEGRATION_TEST_URL), plugin_name),
        headers={"Authorization": f"Bearer {token.value}"},
    )


# For reuse with the protocol code
COMMANDS: list[type[_BaseCommand]] = [
    AssessCommand,
    DiagnoseCommand,
    GoalCommand,
    HistoryOfPresentIllnessCommand,
    MedicationStatementCommand,
    PlanCommand,
    PrescribeCommand,
    QuestionnaireCommand,
    ReasonForVisitCommand,
    StopMedicationCommand,
    UpdateGoalCommand,
]


def create_new_note(token: MaskedValue) -> dict:
    """Create a new note."""
    headers = {
        "Authorization": f"Bearer {token.value}",
        "Content-Type": "application/json",
        "Accept": "application/json",
    }
    data = {
        "patient": 1,
        "provider": 1,
        "note_type": "office",
        "note_type_version": 1,
        "lastModifiedBySessionKey": "8fee3c03a525cebee1d8a6b8e63dd4dg",
    }
    return requests.post(
        f"{settings.INTEGRATION_TEST_URL}/api/Note/", headers=headers, json=data
    ).json()


def get_token() -> MaskedValue:
    """Get a valid token."""
    return MaskedValue(
        requests.post(
            f"{settings.INTEGRATION_TEST_URL}/auth/token/",
            headers={"Content-Type": "application/x-www-form-urlencoded"},
            data={
                "grant_type": "client_credentials",
                "client_id": settings.INTEGRATION_TEST_CLIENT_ID,
                "client_secret": settings.INTEGRATION_TEST_CLIENT_SECRET,
            },
        ).json()["access_token"]
    )



from google.protobuf.internal import enum_type_wrapper as _enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class EffectType(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = ()
    UNKNOWN_EFFECT: _ClassVar[EffectType]
    LOG: _ClassVar[EffectType]
    ADD_PLAN_COMMAND: _ClassVar[EffectType]
    AUTOCOMPLETE_SEARCH_RESULTS: _ClassVar[EffectType]
    ADD_BANNER_ALERT: _ClassVar[EffectType]
    REMOVE_BANNER_ALERT: _ClassVar[EffectType]
    ORIGINATE_ASSESS_COMMAND: _ClassVar[EffectType]
    EDIT_ASSESS_COMMAND: _ClassVar[EffectType]
    DELETE_ASSESS_COMMAND: _ClassVar[EffectType]
    COMMIT_ASSESS_COMMAND: _ClassVar[EffectType]
    ENTER_IN_ERROR_ASSESS_COMMAND: _ClassVar[EffectType]
    ORIGINATE_DIAGNOSE_COMMAND: _ClassVar[EffectType]
    EDIT_DIAGNOSE_COMMAND: _ClassVar[EffectType]
    DELETE_DIAGNOSE_COMMAND: _ClassVar[EffectType]
    COMMIT_DIAGNOSE_COMMAND: _ClassVar[EffectType]
    ENTER_IN_ERROR_DIAGNOSE_COMMAND: _ClassVar[EffectType]
    ORIGINATE_GOAL_COMMAND: _ClassVar[EffectType]
    EDIT_GOAL_COMMAND: _ClassVar[EffectType]
    DELETE_GOAL_COMMAND: _ClassVar[EffectType]
    COMMIT_GOAL_COMMAND: _ClassVar[EffectType]
    ENTER_IN_ERROR_GOAL_COMMAND: _ClassVar[EffectType]
    ORIGINATE_HPI_COMMAND: _ClassVar[EffectType]
    EDIT_HPI_COMMAND: _ClassVar[EffectType]
    DELETE_HPI_COMMAND: _ClassVar[EffectType]
    COMMIT_HPI_COMMAND: _ClassVar[EffectType]
    ENTER_IN_ERROR_HPI_COMMAND: _ClassVar[EffectType]
    ORIGINATE_MEDICATION_STATEMENT_COMMAND: _ClassVar[EffectType]
    EDIT_MEDICATION_STATEMENT_COMMAND: _ClassVar[EffectType]
    DELETE_MEDICATION_STATEMENT_COMMAND: _ClassVar[EffectType]
    COMMIT_MEDICATION_STATEMENT_COMMAND: _ClassVar[EffectType]
    ENTER_IN_ERROR_MEDICATION_STATEMENT_COMMAND: _ClassVar[EffectType]
    ORIGINATE_PLAN_COMMAND: _ClassVar[EffectType]
    EDIT_PLAN_COMMAND: _ClassVar[EffectType]
    DELETE_PLAN_COMMAND: _ClassVar[EffectType]
    COMMIT_PLAN_COMMAND: _ClassVar[EffectType]
    ENTER_IN_ERROR_PLAN_COMMAND: _ClassVar[EffectType]
    ORIGINATE_PRESCRIBE_COMMAND: _ClassVar[EffectType]
    EDIT_PRESCRIBE_COMMAND: _ClassVar[EffectType]
    DELETE_PRESCRIBE_COMMAND: _ClassVar[EffectType]
    COMMIT_PRESCRIBE_COMMAND: _ClassVar[EffectType]
    ENTER_IN_ERROR_PRESCRIBE_COMMAND: _ClassVar[EffectType]
    ORIGINATE_QUESTIONNAIRE_COMMAND: _ClassVar[EffectType]
    EDIT_QUESTIONNAIRE_COMMAND: _ClassVar[EffectType]
    DELETE_QUESTIONNAIRE_COMMAND: _ClassVar[EffectType]
    COMMIT_QUESTIONNAIRE_COMMAND: _ClassVar[EffectType]
    ENTER_IN_ERROR_QUESTIONNAIRE_COMMAND: _ClassVar[EffectType]
    ORIGINATE_REASON_FOR_VISIT_COMMAND: _ClassVar[EffectType]
    EDIT_REASON_FOR_VISIT_COMMAND: _ClassVar[EffectType]
    DELETE_REASON_FOR_VISIT_COMMAND: _ClassVar[EffectType]
    COMMIT_REASON_FOR_VISIT_COMMAND: _ClassVar[EffectType]
    ENTER_IN_ERROR_REASON_FOR_VISIT_COMMAND: _ClassVar[EffectType]
    ORIGINATE_STOP_MEDICATION_COMMAND: _ClassVar[EffectType]
    EDIT_STOP_MEDICATION_COMMAND: _ClassVar[EffectType]
    DELETE_STOP_MEDICATION_COMMAND: _ClassVar[EffectType]
    COMMIT_STOP_MEDICATION_COMMAND: _ClassVar[EffectType]
    ENTER_IN_ERROR_STOP_MEDICATION_COMMAND: _ClassVar[EffectType]
    ORIGINATE_UPDATE_GOAL_COMMAND: _ClassVar[EffectType]
    EDIT_UPDATE_GOAL_COMMAND: _ClassVar[EffectType]
    DELETE_UPDATE_GOAL_COMMAND: _ClassVar[EffectType]
    COMMIT_UPDATE_GOAL_COMMAND: _ClassVar[EffectType]
    ENTER_IN_ERROR_UPDATE_GOAL_COMMAND: _ClassVar[EffectType]
    ORIGINATE_PERFORM_COMMAND: _ClassVar[EffectType]
    EDIT_PERFORM_COMMAND: _ClassVar[EffectType]
    DELETE_PERFORM_COMMAND: _ClassVar[EffectType]
    COMMIT_PERFORM_COMMAND: _ClassVar[EffectType]
    ENTER_IN_ERROR_PERFORM_COMMAND: _ClassVar[EffectType]
    ORIGINATE_INSTRUCT_COMMAND: _ClassVar[EffectType]
    EDIT_INSTRUCT_COMMAND: _ClassVar[EffectType]
    DELETE_INSTRUCT_COMMAND: _ClassVar[EffectType]
    COMMIT_INSTRUCT_COMMAND: _ClassVar[EffectType]
    ENTER_IN_ERROR_INSTRUCT_COMMAND: _ClassVar[EffectType]
    ORIGINATE_LAB_ORDER_COMMAND: _ClassVar[EffectType]
    EDIT_LAB_ORDER_COMMAND: _ClassVar[EffectType]
    DELETE_LAB_ORDER_COMMAND: _ClassVar[EffectType]
    COMMIT_LAB_ORDER_COMMAND: _ClassVar[EffectType]
    ENTER_IN_ERROR_LAB_ORDER_COMMAND: _ClassVar[EffectType]
    ORIGINATE_FAMILY_HISTORY_COMMAND: _ClassVar[EffectType]
    EDIT_FAMILY_HISTORY_COMMAND: _ClassVar[EffectType]
    DELETE_FAMILY_HISTORY_COMMAND: _ClassVar[EffectType]
    COMMIT_FAMILY_HISTORY_COMMAND: _ClassVar[EffectType]
    ENTER_IN_ERROR_FAMILY_HISTORY_COMMAND: _ClassVar[EffectType]
    ORIGINATE_ALLERGY_COMMAND: _ClassVar[EffectType]
    EDIT_ALLERGY_COMMAND: _ClassVar[EffectType]
    DELETE_ALLERGY_COMMAND: _ClassVar[EffectType]
    COMMIT_ALLERGY_COMMAND: _ClassVar[EffectType]
    ENTER_IN_ERROR_ALLERGY_COMMAND: _ClassVar[EffectType]
    ORIGINATE_REMOVE_ALLERGY_COMMAND: _ClassVar[EffectType]
    EDIT_REMOVE_ALLERGY_COMMAND: _ClassVar[EffectType]
    DELETE_REMOVE_ALLERGY_COMMAND: _ClassVar[EffectType]
    COMMIT_REMOVE_ALLERGY_COMMAND: _ClassVar[EffectType]
    ENTER_IN_ERROR_REMOVE_ALLERGY_COMMAND: _ClassVar[EffectType]
    ORIGINATE_SURGICAL_HISTORY_COMMAND: _ClassVar[EffectType]
    EDIT_SURGICAL_HISTORY_COMMAND: _ClassVar[EffectType]
    DELETE_SURGICAL_HISTORY_COMMAND: _ClassVar[EffectType]
    COMMIT_SURGICAL_HISTORY_COMMAND: _ClassVar[EffectType]
    ENTER_IN_ERROR_SURGICAL_HISTORY_COMMAND: _ClassVar[EffectType]
    CREATE_TASK: _ClassVar[EffectType]
    UPDATE_TASK: _ClassVar[EffectType]
    CREATE_TASK_COMMENT: _ClassVar[EffectType]
    ORIGINATE_MEDICAL_HISTORY_COMMAND: _ClassVar[EffectType]
    EDIT_MEDICAL_HISTORY_COMMAND: _ClassVar[EffectType]
    DELETE_MEDICAL_HISTORY_COMMAND: _ClassVar[EffectType]
    COMMIT_MEDICAL_HISTORY_COMMAND: _ClassVar[EffectType]
    ENTER_IN_ERROR_MEDICAL_HISTORY_COMMAND: _ClassVar[EffectType]
    ADD_OR_UPDATE_PROTOCOL_CARD: _ClassVar[EffectType]
    ORIGINATE_TASK_COMMAND: _ClassVar[EffectType]
    EDIT_TASK_COMMAND: _ClassVar[EffectType]
    DELETE_TASK_COMMAND: _ClassVar[EffectType]
    COMMIT_TASK_COMMAND: _ClassVar[EffectType]
    ENTER_IN_ERROR_TASK_COMMAND: _ClassVar[EffectType]
    ORIGINATE_REFILL_COMMAND: _ClassVar[EffectType]
    EDIT_REFILL_COMMAND: _ClassVar[EffectType]
    DELETE_REFILL_COMMAND: _ClassVar[EffectType]
    COMMIT_REFILL_COMMAND: _ClassVar[EffectType]
    ENTER_IN_ERROR_REFILL_COMMAND: _ClassVar[EffectType]
    ORIGINATE_VITALS_COMMAND: _ClassVar[EffectType]
    EDIT_VITALS_COMMAND: _ClassVar[EffectType]
    DELETE_VITALS_COMMAND: _ClassVar[EffectType]
    COMMIT_VITALS_COMMAND: _ClassVar[EffectType]
    ENTER_IN_ERROR_VITALS_COMMAND: _ClassVar[EffectType]
    ORIGINATE_UPDATE_DIAGNOSIS_COMMAND: _ClassVar[EffectType]
    EDIT_UPDATE_DIAGNOSIS_COMMAND: _ClassVar[EffectType]
    DELETE_UPDATE_DIAGNOSIS_COMMAND: _ClassVar[EffectType]
    COMMIT_UPDATE_DIAGNOSIS_COMMAND: _ClassVar[EffectType]
    ENTER_IN_ERROR_UPDATE_DIAGNOSIS_COMMAND: _ClassVar[EffectType]
    ORIGINATE_CLOSE_GOAL_COMMAND: _ClassVar[EffectType]
    EDIT_CLOSE_GOAL_COMMAND: _ClassVar[EffectType]
    DELETE_CLOSE_GOAL_COMMAND: _ClassVar[EffectType]
    COMMIT_CLOSE_GOAL_COMMAND: _ClassVar[EffectType]
    ENTER_IN_ERROR_CLOSE_GOAL_COMMAND: _ClassVar[EffectType]
    ANNOTATE_PATIENT_CHART_CONDITION_RESULTS: _ClassVar[EffectType]
    ANNOTATE_CLAIM_CONDITION_RESULTS: _ClassVar[EffectType]
    SHOW_PATIENT_CHART_SUMMARY_SECTIONS: _ClassVar[EffectType]
    SHOW_PATIENT_PROFILE_SECTIONS: _ClassVar[EffectType]
    PATIENT_PROFILE__ADD_PHARMACY__POST_SEARCH_RESULTS: _ClassVar[EffectType]
    SEND_SURESCRIPTS_ELIGIBILITY_REQUEST: _ClassVar[EffectType]
    SEND_SURESCRIPTS_MEDICATION_HISTORY_REQUEST: _ClassVar[EffectType]
    SEND_SURESCRIPTS_BENEFITS_REQUEST: _ClassVar[EffectType]
    ORIGINATE_EXAM_COMMAND: _ClassVar[EffectType]
    EDIT_EXAM_COMMAND: _ClassVar[EffectType]
    DELETE_EXAM_COMMAND: _ClassVar[EffectType]
    COMMIT_EXAM_COMMAND: _ClassVar[EffectType]
    ENTER_IN_ERROR_EXAM_COMMAND: _ClassVar[EffectType]
    ORIGINATE_ROS_COMMAND: _ClassVar[EffectType]
    EDIT_ROS_COMMAND: _ClassVar[EffectType]
    DELETE_ROS_COMMAND: _ClassVar[EffectType]
    COMMIT_ROS_COMMAND: _ClassVar[EffectType]
    ENTER_IN_ERROR_ROS_COMMAND: _ClassVar[EffectType]
    ORIGINATE_STRUCTURED_ASSESSMENT_COMMAND: _ClassVar[EffectType]
    EDIT_STRUCTURED_ASSESSMENT_COMMAND: _ClassVar[EffectType]
    DELETE_STRUCTURED_ASSESSMENT_COMMAND: _ClassVar[EffectType]
    COMMIT_STRUCTURED_ASSESSMENT_COMMAND: _ClassVar[EffectType]
    ENTER_IN_ERROR_STRUCTURED_ASSESSMENT_COMMAND: _ClassVar[EffectType]
    SHOW_ACTION_BUTTON: _ClassVar[EffectType]
    PATIENT_PORTAL__INTAKE_FORM_RESULTS: _ClassVar[EffectType]
    LAUNCH_MODAL: _ClassVar[EffectType]
UNKNOWN_EFFECT: EffectType
LOG: EffectType
ADD_PLAN_COMMAND: EffectType
AUTOCOMPLETE_SEARCH_RESULTS: EffectType
ADD_BANNER_ALERT: EffectType
REMOVE_BANNER_ALERT: EffectType
ORIGINATE_ASSESS_COMMAND: EffectType
EDIT_ASSESS_COMMAND: EffectType
DELETE_ASSESS_COMMAND: EffectType
COMMIT_ASSESS_COMMAND: EffectType
ENTER_IN_ERROR_ASSESS_COMMAND: EffectType
ORIGINATE_DIAGNOSE_COMMAND: EffectType
EDIT_DIAGNOSE_COMMAND: EffectType
DELETE_DIAGNOSE_COMMAND: EffectType
COMMIT_DIAGNOSE_COMMAND: EffectType
ENTER_IN_ERROR_DIAGNOSE_COMMAND: EffectType
ORIGINATE_GOAL_COMMAND: EffectType
EDIT_GOAL_COMMAND: EffectType
DELETE_GOAL_COMMAND: EffectType
COMMIT_GOAL_COMMAND: EffectType
ENTER_IN_ERROR_GOAL_COMMAND: EffectType
ORIGINATE_HPI_COMMAND: EffectType
EDIT_HPI_COMMAND: EffectType
DELETE_HPI_COMMAND: EffectType
COMMIT_HPI_COMMAND: EffectType
ENTER_IN_ERROR_HPI_COMMAND: EffectType
ORIGINATE_MEDICATION_STATEMENT_COMMAND: EffectType
EDIT_MEDICATION_STATEMENT_COMMAND: EffectType
DELETE_MEDICATION_STATEMENT_COMMAND: EffectType
COMMIT_MEDICATION_STATEMENT_COMMAND: EffectType
ENTER_IN_ERROR_MEDICATION_STATEMENT_COMMAND: EffectType
ORIGINATE_PLAN_COMMAND: EffectType
EDIT_PLAN_COMMAND: EffectType
DELETE_PLAN_COMMAND: EffectType
COMMIT_PLAN_COMMAND: EffectType
ENTER_IN_ERROR_PLAN_COMMAND: EffectType
ORIGINATE_PRESCRIBE_COMMAND: EffectType
EDIT_PRESCRIBE_COMMAND: EffectType
DELETE_PRESCRIBE_COMMAND: EffectType
COMMIT_PRESCRIBE_COMMAND: EffectType
ENTER_IN_ERROR_PRESCRIBE_COMMAND: EffectType
ORIGINATE_QUESTIONNAIRE_COMMAND: EffectType
EDIT_QUESTIONNAIRE_COMMAND: EffectType
DELETE_QUESTIONNAIRE_COMMAND: EffectType
COMMIT_QUESTIONNAIRE_COMMAND: EffectType
ENTER_IN_ERROR_QUESTIONNAIRE_COMMAND: EffectType
ORIGINATE_REASON_FOR_VISIT_COMMAND: EffectType
EDIT_REASON_FOR_VISIT_COMMAND: EffectType
DELETE_REASON_FOR_VISIT_COMMAND: EffectType
COMMIT_REASON_FOR_VISIT_COMMAND: EffectType
ENTER_IN_ERROR_REASON_FOR_VISIT_COMMAND: EffectType
ORIGINATE_STOP_MEDICATION_COMMAND: EffectType
EDIT_STOP_MEDICATION_COMMAND: EffectType
DELETE_STOP_MEDICATION_COMMAND: EffectType
COMMIT_STOP_MEDICATION_COMMAND: EffectType
ENTER_IN_ERROR_STOP_MEDICATION_COMMAND: EffectType
ORIGINATE_UPDATE_GOAL_COMMAND: EffectType
EDIT_UPDATE_GOAL_COMMAND: EffectType
DELETE_UPDATE_GOAL_COMMAND: EffectType
COMMIT_UPDATE_GOAL_COMMAND: EffectType
ENTER_IN_ERROR_UPDATE_GOAL_COMMAND: EffectType
ORIGINATE_PERFORM_COMMAND: EffectType
EDIT_PERFORM_COMMAND: EffectType
DELETE_PERFORM_COMMAND: EffectType
COMMIT_PERFORM_COMMAND: EffectType
ENTER_IN_ERROR_PERFORM_COMMAND: EffectType
ORIGINATE_INSTRUCT_COMMAND: EffectType
EDIT_INSTRUCT_COMMAND: EffectType
DELETE_INSTRUCT_COMMAND: EffectType
COMMIT_INSTRUCT_COMMAND: EffectType
ENTER_IN_ERROR_INSTRUCT_COMMAND: EffectType
ORIGINATE_LAB_ORDER_COMMAND: EffectType
EDIT_LAB_ORDER_COMMAND: EffectType
DELETE_LAB_ORDER_COMMAND: EffectType
COMMIT_LAB_ORDER_COMMAND: EffectType
ENTER_IN_ERROR_LAB_ORDER_COMMAND: EffectType
ORIGINATE_FAMILY_HISTORY_COMMAND: EffectType
EDIT_FAMILY_HISTORY_COMMAND: EffectType
DELETE_FAMILY_HISTORY_COMMAND: EffectType
COMMIT_FAMILY_HISTORY_COMMAND: EffectType
ENTER_IN_ERROR_FAMILY_HISTORY_COMMAND: EffectType
ORIGINATE_ALLERGY_COMMAND: EffectType
EDIT_ALLERGY_COMMAND: EffectType
DELETE_ALLERGY_COMMAND: EffectType
COMMIT_ALLERGY_COMMAND: EffectType
ENTER_IN_ERROR_ALLERGY_COMMAND: EffectType
ORIGINATE_REMOVE_ALLERGY_COMMAND: EffectType
EDIT_REMOVE_ALLERGY_COMMAND: EffectType
DELETE_REMOVE_ALLERGY_COMMAND: EffectType
COMMIT_REMOVE_ALLERGY_COMMAND: EffectType
ENTER_IN_ERROR_REMOVE_ALLERGY_COMMAND: EffectType
ORIGINATE_SURGICAL_HISTORY_COMMAND: EffectType
EDIT_SURGICAL_HISTORY_COMMAND: EffectType
DELETE_SURGICAL_HISTORY_COMMAND: EffectType
COMMIT_SURGICAL_HISTORY_COMMAND: EffectType
ENTER_IN_ERROR_SURGICAL_HISTORY_COMMAND: EffectType
CREATE_TASK: EffectType
UPDATE_TASK: EffectType
CREATE_TASK_COMMENT: EffectType
ORIGINATE_MEDICAL_HISTORY_COMMAND: EffectType
EDIT_MEDICAL_HISTORY_COMMAND: EffectType
DELETE_MEDICAL_HISTORY_COMMAND: EffectType
COMMIT_MEDICAL_HISTORY_COMMAND: EffectType
ENTER_IN_ERROR_MEDICAL_HISTORY_COMMAND: EffectType
ADD_OR_UPDATE_PROTOCOL_CARD: EffectType
ORIGINATE_TASK_COMMAND: EffectType
EDIT_TASK_COMMAND: EffectType
DELETE_TASK_COMMAND: EffectType
COMMIT_TASK_COMMAND: EffectType
ENTER_IN_ERROR_TASK_COMMAND: EffectType
ORIGINATE_REFILL_COMMAND: EffectType
EDIT_REFILL_COMMAND: EffectType
DELETE_REFILL_COMMAND: EffectType
COMMIT_REFILL_COMMAND: EffectType
ENTER_IN_ERROR_REFILL_COMMAND: EffectType
ORIGINATE_VITALS_COMMAND: EffectType
EDIT_VITALS_COMMAND: EffectType
DELETE_VITALS_COMMAND: EffectType
COMMIT_VITALS_COMMAND: EffectType
ENTER_IN_ERROR_VITALS_COMMAND: EffectType
ORIGINATE_UPDATE_DIAGNOSIS_COMMAND: EffectType
EDIT_UPDATE_DIAGNOSIS_COMMAND: EffectType
DELETE_UPDATE_DIAGNOSIS_COMMAND: EffectType
COMMIT_UPDATE_DIAGNOSIS_COMMAND: EffectType
ENTER_IN_ERROR_UPDATE_DIAGNOSIS_COMMAND: EffectType
ORIGINATE_CLOSE_GOAL_COMMAND: EffectType
EDIT_CLOSE_GOAL_COMMAND: EffectType
DELETE_CLOSE_GOAL_COMMAND: EffectType
COMMIT_CLOSE_GOAL_COMMAND: EffectType
ENTER_IN_ERROR_CLOSE_GOAL_COMMAND: EffectType
ANNOTATE_PATIENT_CHART_CONDITION_RESULTS: EffectType
ANNOTATE_CLAIM_CONDITION_RESULTS: EffectType
SHOW_PATIENT_CHART_SUMMARY_SECTIONS: EffectType
SHOW_PATIENT_PROFILE_SECTIONS: EffectType
PATIENT_PROFILE__ADD_PHARMACY__POST_SEARCH_RESULTS: EffectType
SEND_SURESCRIPTS_ELIGIBILITY_REQUEST: EffectType
SEND_SURESCRIPTS_MEDICATION_HISTORY_REQUEST: EffectType
SEND_SURESCRIPTS_BENEFITS_REQUEST: EffectType
ORIGINATE_EXAM_COMMAND: EffectType
EDIT_EXAM_COMMAND: EffectType
DELETE_EXAM_COMMAND: EffectType
COMMIT_EXAM_COMMAND: EffectType
ENTER_IN_ERROR_EXAM_COMMAND: EffectType
ORIGINATE_ROS_COMMAND: EffectType
EDIT_ROS_COMMAND: EffectType
DELETE_ROS_COMMAND: EffectType
COMMIT_ROS_COMMAND: EffectType
ENTER_IN_ERROR_ROS_COMMAND: EffectType
ORIGINATE_STRUCTURED_ASSESSMENT_COMMAND: EffectType
EDIT_STRUCTURED_ASSESSMENT_COMMAND: EffectType
DELETE_STRUCTURED_ASSESSMENT_COMMAND: EffectType
COMMIT_STRUCTURED_ASSESSMENT_COMMAND: EffectType
ENTER_IN_ERROR_STRUCTURED_ASSESSMENT_COMMAND: EffectType
SHOW_ACTION_BUTTON: EffectType
PATIENT_PORTAL__INTAKE_FORM_RESULTS: EffectType
LAUNCH_MODAL: EffectType

class Effect(_message.Message):
    __slots__ = ("type", "payload", "plugin_name", "classname")
    TYPE_FIELD_NUMBER: _ClassVar[int]
    PAYLOAD_FIELD_NUMBER: _ClassVar[int]
    PLUGIN_NAME_FIELD_NUMBER: _ClassVar[int]
    CLASSNAME_FIELD_NUMBER: _ClassVar[int]
    type: EffectType
    payload: str
    plugin_name: str
    classname: str
    def __init__(self, type: _Optional[_Union[EffectType, str]] = ..., payload: _Optional[str] = ..., plugin_name: _Optional[str] = ..., classname: _Optional[str] = ...) -> None: ...
